from __future__ import absolute_import, division, print_function, unicode_literals
import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
from ...utils.energy import EnergyBounds
from ...irf import EnergyDispersion, EffectiveAreaTable
from .. import SpectrumExtraction, SpectrumSimulation
from ..models import PowerLaw


class TestSpectrumSimulation:
    def setup(self):
        e_true = SpectrumExtraction.DEFAULT_TRUE_ENERGY
        e_reco = SpectrumExtraction.DEFAULT_RECO_ENERGY

        edisp = EnergyDispersion.from_gauss(
            e_true=e_true, e_reco=e_reco, sigma=0.2, bias=0
        )

        aeff = EffectiveAreaTable.from_parametrization(energy=e_true)

        self.source_model = PowerLaw(
            index=2.3, amplitude="2.5e-12 cm-2 s-1 TeV-1", reference="1 TeV"
        )
        self.background_model = PowerLaw(
            index=3, amplitude="3e-12 cm-2 s-1 TeV-1", reference="1 TeV"
        )
        self.alpha = 1.0 / 3

        # Minimal setup
        self.sim = SpectrumSimulation(
            aeff=aeff, edisp=edisp, source_model=self.source_model, livetime=4 * u.h
        )

    def test_without_background(self):
        self.sim.simulate_obs(seed=23, obs_id=23)
        assert self.sim.obs.on_vector.total_counts == 160

    def test_with_background(self):
        self.sim.background_model = self.background_model
        self.sim.alpha = self.alpha
        self.sim.simulate_obs(seed=23, obs_id=23)
        assert self.sim.obs.on_vector.total_counts == 530
        assert self.sim.obs.off_vector.total_counts == 1112

    def test_observations_list(self):
        seeds = np.arange(5)
        self.sim.run(seed=seeds)
        assert (self.sim.result.obs_id == seeds).all()
        assert self.sim.result[0].on_vector.total_counts == 158
        assert self.sim.result[1].on_vector.total_counts == 158
        assert self.sim.result[2].on_vector.total_counts == 161
        assert self.sim.result[3].on_vector.total_counts == 168
        assert self.sim.result[4].on_vector.total_counts == 186

    def test_without_edisp(self):
        sim = SpectrumSimulation(
            aeff=self.sim.aeff, source_model=self.sim.source_model, livetime=4 * u.h
        )
        sim.simulate_obs(seed=23, obs_id=23)
        assert sim.obs.on_vector.total_counts == 161
        # The test value is taken from the test with edisp
        assert_allclose(
            np.sum(sim.npred_source.data.data.value), 167.467572145, rtol=0.01
        )

    def test_without_aeff(self):
        e_true = EnergyBounds.equal_log_spacing(1, 10, 5, u.TeV)
        self.source_model.parameters["amplitude"].quantity = "1 TeV-1 s-1"
        sim = SpectrumSimulation(
            source_model=self.source_model, livetime=4 * u.h, e_true=e_true
        )
        sim.simulate_obs(seed=23, obs_id=23)
        assert sim.obs.on_vector.total_counts == 10509



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import logging
import numpy as np
import astropy.units as u
from regions import CircleSkyRegion
from ..utils.scripts import make_path
from ..irf import PSF3D
from .core import PHACountsSpectrum
from .observation import SpectrumObservation, SpectrumObservationList

__all__ = ["SpectrumExtraction"]

log = logging.getLogger(__name__)


class SpectrumExtraction(object):
    """Creating input data to 1D spectrum fitting.

    This class is responsible for extracting a
    `~gammapy.spectrum.SpectrumObservation` from a
    `~gammapy.data.DataStoreObservation`. The background estimation is done
    beforehand, using e.g. the
    `~gammapy.background.ReflectedRegionsBackgroundEstimator`. For point
    sources analyzed with 'full containment' IRFs, a correction for PSF
    leakage out of the circular ON region can be applied.

    For more info see :ref:`spectral_fitting`.

    For a usage example see :gp-notebook:`spectrum_analysis`

    Parameters
    ----------
    observations : `~gammapy.data.Observations`
        Observations to process
    bkg_estimate : `~gammapy.background.BackgroundEstimate`
        Background estimate, e.g. of
        `~gammapy.background.ReflectedRegionsBackgroundEstimator`
    e_reco : `~astropy.units.Quantity`, optional
        Reconstructed energy binning
    e_true : `~astropy.units.Quantity`, optional
        True energy binning
    containment_correction : bool
        Apply containment correction for point sources and circular on regions.
    max_alpha : float
        Maximum alpha value to accept, if the background was estimated using
        reflected regions this is 1 / minimum number of regions.
    use_recommended_erange : bool
        Extract spectrum only within the recommended valid energy range of the
        effective area table (default is True).
    """

    DEFAULT_TRUE_ENERGY = np.logspace(-2, 2.5, 109) * u.TeV
    """True energy axis to be used if not specified otherwise"""
    DEFAULT_RECO_ENERGY = np.logspace(-2, 2, 73) * u.TeV
    """Reconstruced energy axis to be used if not specified otherwise"""

    def __init__(
        self,
        observations,
        bkg_estimate,
        e_reco=None,
        e_true=None,
        containment_correction=False,
        max_alpha=1,
        use_recommended_erange=True,
    ):

        self.observations = observations
        self.bkg_estimate = bkg_estimate
        self.e_reco = e_reco if e_reco is not None else self.DEFAULT_RECO_ENERGY
        self.e_true = e_true if e_true is not None else self.DEFAULT_TRUE_ENERGY
        self.containment_correction = containment_correction
        self.max_alpha = max_alpha
        self.use_recommended_erange = use_recommended_erange
        self.spectrum_observations = SpectrumObservationList()

        self.containment = None
        self._on_vector = None
        self._off_vector = None
        self._aeff = None
        self._edisp = None

    def run(self):
        """Run all steps.
        """
        log.info("Running {}".format(self))
        for obs, bkg in zip(self.observations, self.bkg_estimate):
            if not self._alpha_ok(bkg):
                continue
            self.spectrum_observations.append(self.process(obs, bkg))

    def _alpha_ok(self, bkg):
        """Check if observation fulfills alpha criterion"""
        condition = bkg.a_off == 0 or bkg.a_on / bkg.a_off > self.max_alpha
        if condition:
            msg = "Skipping because {} / {} > {}"
            log.info(msg.format(bkg.a_on, bkg.a_off, self.max_alpha))
            return False
        else:
            return True

    def process(self, observation, bkg):
        """Process one observation.

        Parameters
        ----------
        observation : `~gammapy.data.DataStoreObservation`
            Observation
        bkg : `~gammapy.background.BackgroundEstimate`
            Background estimate

        Returns
        -------
        spectrum_observation : `~gammapy.spectrum.SpectrumObservation`
            Spectrum observation
        """
        log.info("Process observation\n {}".format(observation))
        self.make_empty_vectors(observation, bkg)
        self.extract_counts(bkg)
        self.extract_irfs(observation)

        if self.containment_correction:
            self.apply_containment_correction(observation, bkg)
        else:
            self.containment = np.ones(self._aeff.energy.nbins)

        spectrum_observation = SpectrumObservation(
            on_vector=self._on_vector,
            aeff=self._aeff,
            off_vector=self._off_vector,
            edisp=self._edisp,
        )

        if self.use_recommended_erange:
            try:
                spectrum_observation.hi_threshold = observation.aeff.high_threshold
                spectrum_observation.lo_threshold = observation.aeff.low_threshold
            except KeyError:
                log.warning("No thresholds defined for obs {}".format(observation))

        return spectrum_observation

    def make_empty_vectors(self, observation, bkg):
        """Create empty vectors.

        This method copies over all meta info and sets up the energy binning.

        Parameters
        ----------
        observation : `~gammapy.data.DataStoreObservation`
            Observation
        bkg : `~gammapy.background.BackgroundEstimate`
            Background estimate
        """
        log.info("Update observation meta info")

        offset = observation.pointing_radec.separation(bkg.on_region.center)
        log.info("Offset : {}\n".format(offset))

        self._on_vector = PHACountsSpectrum(
            energy_lo=self.e_reco[:-1],
            energy_hi=self.e_reco[1:],
            backscal=bkg.a_on,
            offset=offset,
            livetime=observation.observation_live_time_duration,
            obs_id=observation.obs_id,
        )

        self._off_vector = self._on_vector.copy()
        self._off_vector.is_bkg = True
        self._off_vector.backscal = bkg.a_off

    def extract_counts(self, bkg):
        """Fill on and off vector for one observation.

        Parameters
        ----------
        bkg : `~gammapy.background.BackgroundEstimate`
            Background estimate
        """
        log.info("Fill events")
        self._on_vector.fill(bkg.on_events)
        self._off_vector.fill(bkg.off_events)

    def extract_irfs(self, observation):
        """Extract IRFs.

        Parameters
        ----------
        observation : `~gammapy.data.DataStoreObservation`
            Observation
        """
        log.info("Extract IRFs")
        offset = self._on_vector.offset
        self._aeff = observation.aeff.to_effective_area_table(
            offset, energy=self.e_true
        )
        self._edisp = observation.edisp.to_energy_dispersion(
            offset, e_reco=self.e_reco, e_true=self.e_true
        )

    def apply_containment_correction(self, observation, bkg):
        """Apply PSF containment correction.

        Parameters
        ----------
        observation : `~gammapy.data.DataStoreObservation`
            observation
        bkg : `~gammapy.background.BackgroundEstimate`
            background esimate
        """
        # TODO: This should be split out into a separate class
        if not isinstance(bkg.on_region, CircleSkyRegion):
            raise TypeError(
                "Incorrect region type for containment correction."
                " Should be CircleSkyRegion."
            )

        log.info("Apply containment correction")
        # First need psf
        angles = np.linspace(0.0, 1.5, 150) * u.deg
        offset = self._on_vector.offset
        if isinstance(observation.psf, PSF3D):
            psf = observation.psf.to_energy_dependent_table_psf(theta=offset)
        else:
            psf = observation.psf.to_energy_dependent_table_psf(offset, angles)

        center_energies = self._aeff.energy.nodes
        containment = []
        for index, energy in enumerate(center_energies):
            try:
                cont_ = psf.integral(energy, 0.0 * u.deg, bkg.on_region.radius)
            except:
                msg = "Containment correction failed for bin {}, energy {}."
                log.warning(msg.format(index, energy))
                cont_ = 1
            finally:
                containment.append(cont_)

        self.containment = np.array(containment)
        self._aeff.data.data *= self.containment

    def compute_energy_threshold(self, **kwargs):
        """Compute and set the safe energy threshold for all observations.

        See `SpectrumObservation.compute_energy_threshold` for full
        documentation about the options.
        """

        for obs in self.spectrum_observations:
            obs.compute_energy_threshold(**kwargs)

    def write(self, outdir, ogipdir="ogip_data", use_sherpa=False, overwrite=False):
        """Write results to disk.

        Parameters
        ----------
        outdir : `~gammapy.extern.pathlib.Path`
            Output folder
        ogipdir : str, optional
            Folder name for OGIP data, default: 'ogip_data'
        use_sherpa : bool, optional
            Write Sherpa compliant files?
        overwrite : bool
            Overwrite existing files?
        """
        outdir = make_path(outdir)
        log.info("Writing OGIP files to {}".format(outdir / ogipdir))
        outdir.mkdir(exist_ok=True, parents=True)
        self.spectrum_observations.write(
            outdir / ogipdir, use_sherpa=use_sherpa, overwrite=overwrite
        )

        # TODO : add more debug plots etc. here



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from ...data import DataStore
from ...utils.testing import requires_data

# TODO: clean up these FITS production tests
# There is some duplication with `gammapy/data/tests/test_data_store.py`
# This should either be moved there, or to a separate test file
# called "test legacy" or "test high level" or something like that
# The goal would be to not accidentally break support for old files
productions = [
    dict(
        prod="hap-hd-prod2",
        datastore="$GAMMAPY_DATA/tests/hess-crab4-hd-hap-prod2",
        test_obs=23523,
        aeff_ref=267252.7,
        psf_type="psf_3gauss",
        psf_ref=106.310,
        edisp_ref=2.059,
    ),
    dict(
        prod="hess-dl3-dr1",
        datastore="$GAMMAPY_DATA/hess-dl3-dr1",
        test_obs=23523,
        aeff_ref=229490.7,
        psf_type="psf_table",
        psf_ref=237.759,
        edisp_ref=2.247,
    ),
]


class FitsProductionTester:
    def __init__(self, prod):
        self.ref_dict = prod
        self.ds = DataStore.from_dir(prod["datastore"])
        self.ref_energy = 1 * u.TeV
        self.ref_offset = 0.25 * u.deg
        self.ref_rad = np.arange(0, 2, 0.1) * u.deg
        self.ref_migra = 0.95
        self.observation = self.ds.obs(prod["test_obs"])

    def test_all(self):
        self.test_aeff()
        self.test_psf()
        self.test_edisp()

    def test_aeff(self):
        aeff = self.observation.load(hdu_type="aeff", hdu_class="aeff_2d")
        actual = aeff.data.evaluate(energy=self.ref_energy, offset=self.ref_offset)
        desired = self.ref_dict["aeff_ref"]
        assert actual.unit == "m2"
        assert_allclose(actual.value, desired, rtol=1e-3)

    def test_edisp(self):
        edisp = self.observation.load(hdu_type="edisp", hdu_class="edisp_2d")
        actual = edisp.data.evaluate(
            e_true=self.ref_energy, offset=self.ref_offset, migra=self.ref_migra
        )
        desired = self.ref_dict["edisp_ref"]
        assert actual.unit == ""
        assert_allclose(actual.value, desired, rtol=1e-3)

    def test_psf(self):
        psf = self.observation.load(hdu_type="psf", hdu_class=self.ref_dict["psf_type"])
        table_psf = psf.to_energy_dependent_table_psf(
            rad=self.ref_rad, theta=self.ref_offset
        )
        actual = table_psf.evaluate(energy=self.ref_energy)
        desired = self.ref_dict["psf_ref"]
        assert actual.unit == "sr-1"
        assert_allclose(actual.value[0][4], desired, rtol=1e-3)


@pytest.mark.parametrize("prod", productions)
@requires_data("gammapy-data")
def test_fits_prods(prod):
    tester = FitsProductionTester(prod)
    tester.test_all()



"""Demonstrate the artefact that can arise if
   fc_fix_limits is not used."""
import numpy as np
from astropy.table import Table
from gammapy.stats import fc_find_acceptance_interval_poisson

background = 3.5
cl = 0.90
x_bins = np.arange(0, 100)

table = Table()
table["mu"] = [
    0.745,
    0.750,
    0.755,
    1.030,
    1.035,
    1.040,
    1.045,
    1.050,
    1.055,
    1.060,
    1.065,
]
table["x_min"] = 0.0
table["x_max"] = 0.0

for row in table:
    x_min, x_max = fc_find_acceptance_interval_poisson(
        row["mu"], background, x_bins, cl
    )
    row["x_min"] = x_min
    row["x_max"] = x_max

table.pprint()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from astropy import units as u
from numpy.testing import assert_allclose
from ...utils.testing import requires_dependency, requires_data, mpl_plot_check
from ...data.event_list import EventListBase, EventList, EventListLAT


@requires_data("gammapy-data")
class TestEventListBase:
    def setup(self):
        filename = "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
        self.events = EventListBase.read(filename)

    @pytest.mark.parametrize(
        "parameter, band", [("ENERGY", (0.8 * u.TeV, 5.0 * u.TeV))]
    )
    def test_select_parameter(self, parameter, band):
        selected_events = self.events.select_parameter(parameter, band)
        assert np.all(
            (selected_events.table[parameter].quantity >= band[0])
            & (selected_events.table[parameter].quantity < band[1])
        )


@requires_data("gammapy-data")
class TestEventListHESS:
    def setup(self):
        filename = (
            "$GAMMAPY_DATA/tests/unbundled/hess/run_0023037_hard_eventlist.fits.gz"
        )
        self.events = EventList.read(filename)

    def test_basics(self):
        assert "EventList" in str(self.events)

        assert len(self.events.table) == 49
        assert self.events.time[0].iso == "2004-10-14 00:08:39.214"
        assert self.events.radec[0].to_string() == "82.7068 19.8186"
        assert self.events.galactic[0].to_string(precision=2) == "185.96 -7.69"
        assert self.events.altaz[0].to_string() == "46.5793 30.8799"
        assert_allclose(self.events.offset[0].value, 1.904497742652893, rtol=1e-5)
        assert "{:1.5f}".format(self.events.energy[0]) == "11.64355 TeV"

        lon, lat, height = self.events.observatory_earth_location.to_geodetic()
        assert "{:1.5f}".format(lon) == "16.50022 deg"
        assert "{:1.5f}".format(lat) == "-23.27178 deg"
        assert "{:1.5f}".format(height) == "1835.00000 m"

    def test_altaz(self):
        altaz = self.events.altaz
        assert_allclose(altaz[0].az.deg, 46.579258, atol=1e-3)
        assert_allclose(altaz[0].alt.deg, 30.879939, atol=1e-3)

        altaz = self.events.altaz_from_table
        assert_allclose(altaz[0].az.deg, 46.205875, atol=1e-3)
        assert_allclose(altaz[0].alt.deg, 31.200132, atol=1e-3)
        # TODO: add asserts for frame properties

    def test_stack(self):
        event_lists = [self.events] * 3
        stacked_list = EventList.stack(event_lists)
        assert len(stacked_list.table) == 49 * 3

    @requires_dependency("matplotlib")
    def test_plot_time(self):
        with mpl_plot_check():
            self.events.plot_time()

    @requires_dependency("matplotlib")
    def test_plot_energy(self):
        with mpl_plot_check():
            self.events.plot_energy()

    @requires_dependency("matplotlib")
    def test_plot_offset2_distribution(self):
        with mpl_plot_check():
            self.events.plot_offset2_distribution()

    @requires_dependency("matplotlib")
    def test_plot_energy_offset(self):
        with mpl_plot_check():
            self.events.plot_energy_offset()

    @requires_dependency("matplotlib")
    def test_plot_image(self):
        with mpl_plot_check():
            self.events.plot_image()

    @requires_dependency("matplotlib")
    def test_peek(self):
        with mpl_plot_check():
            self.events.peek()


@requires_data("gammapy-data")
class TestEventListFermi:
    def setup(self):
        filename = "$GAMMAPY_DATA/fermi-3fhl-gc/fermi-3fhl-gc-events.fits.gz"
        self.events = EventListLAT.read(filename)

    def test_basics(self):
        assert "EventList" in str(self.events)
        assert len(self.events.table) == 32843

    @requires_dependency("matplotlib")
    def test_plot_image(self):
        with mpl_plot_check():
            self.events.plot_image()


@requires_data("gammapy-data")
class TestEventListChecker:
    def setup(self):
        self.event_list = EventList.read(
            "$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_111140.fits"
        )

    def test_check_all(self):
        records = list(self.event_list.check())
        assert len(records) == 3



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from numpy.testing import assert_allclose
from ...utils.testing import requires_data
from ..core import gammapy_data
from ...datasets import load_poisson_stats_image


@requires_data("gammapy-data")
def test_gammapy_data():
    """Try loading a file from gammapy-data.
    """
    assert gammapy_data.dir.is_dir()


@requires_data("gammapy-data")
def test_load_poisson_stats_image():
    data = load_poisson_stats_image()
    assert data.sum() == 40896

    images = load_poisson_stats_image(extra_info=True)
    refs = dict(counts=40896, model=41000, source=1000, background=40000)
    for name, expected in refs.items():
        assert_allclose(images[name].sum(), expected)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Pulsar wind nebula (PWN) source models."""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from astropy.units import Quantity
from astropy.utils import lazyproperty
import astropy.constants as const
from scipy.optimize import fsolve
from ...extern.validator import validate_physical_type
from ..source import Pulsar, SNRTrueloveMcKee

__all__ = ["PWN"]


class PWN(object):
    """Simple pulsar wind nebula (PWN) evolution model.

    Parameters
    ----------
    pulsar : `~gammapy.astro.source.Pulsar`
        Pulsar model instance.
    snr : `~gammapy.astro.source.SNRTrueloveMcKee`
        SNR model instance
    eta_e : float
        Fraction of energy going into electrons.
    eta_B : float
        Fraction of energy going into magnetic fields.
    age : `~astropy.units.Quantity`
        Age of the PWN.
    morphology : str
        Morphology model of the PWN
    """

    def __init__(
        self,
        pulsar=Pulsar(),
        snr=SNRTrueloveMcKee(),
        eta_e=0.999,
        eta_B=0.001,
        morphology="Gaussian2D",
        age=None,
    ):
        self.pulsar = pulsar
        if not isinstance(snr, SNRTrueloveMcKee):
            raise ValueError("SNR must be instance of SNRTrueloveMcKee")
        self.snr = snr
        self.eta_e = eta_e
        self.eta_B = eta_B
        self.morphology = morphology
        if age is not None:
            validate_physical_type("age", age, "time")
            self.age = age

    def _radius_free_expansion(self, t):
        """Radius at age t during free expansion phase.

        Reference: http://adsabs.harvard.edu/abs/2006ARA%26A..44...17G (Formula 8).
        """
        term1 = (self.snr.e_sn ** 3 * self.pulsar.L_0 ** 2) / (self.snr.m_ejecta ** 5)
        return (1.44 * term1 ** (1.0 / 10) * t ** (6.0 / 5)).cgs

    @lazyproperty
    def _collision_time(self):
        """Time of collision between the PWN and the reverse shock of the SNR.

        Returns
        -------
        t_coll : `~astropy.units.Quantity`
            Time of collision.
        """

        def time_coll(t):
            t = Quantity(t, "yr")
            r_pwn = self._radius_free_expansion(t).to_value("cm")
            r_shock = self.snr.radius_reverse_shock(t).to_value("cm")
            return r_pwn - r_shock

        # 4e3 years is a typical value that works for fsolve
        return Quantity(fsolve(time_coll, 4e3), "yr")

    def radius(self, t=None):
        """Radius of the PWN at age t.

        Reference: http://adsabs.harvard.edu/abs/2006ARA%26A..44...17G (Formula 8).

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.

        Notes
        -----
        During the free expansion phase the radius of the PWN evolves like:

        .. math::

            R_{PWN}(t) = 1.44\\text{pc}\\left(\\frac{E_{SN}^3\\dot{E}_0^2}
            {M_{ej}^5}\\right)^{1/10}t^{6/5}

        After the collision with the reverse shock of the SNR, the radius is
        assumed to be constant (See `~gammapy.astro.source.SNRTrueloveMcKee.radius_reverse_shock`)

        """
        if t is not None:
            validate_physical_type("t", t, "time")
        elif hasattr(self, "age"):
            t = self.age
        else:
            raise ValueError("Need time variable or age attribute.")
        # Radius at time of collision
        r_coll = self._radius_free_expansion(self._collision_time)
        r = np.where(
            t < self._collision_time, self._radius_free_expansion(t).value, r_coll.value
        )
        return Quantity(r, "cm")

    def magnetic_field(self, t=None):
        """Estimate of the magnetic field inside the PWN.

        By assuming that a certain fraction of the spin down energy is
        converted to magnetic field energy an estimation of the magnetic
        field can be derived.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        if t is not None:
            validate_physical_type("t", t, "time")
        elif hasattr(self, "age"):
            t = self.age
        else:
            raise ValueError("Need time variable or age attribute.")

        energy = self.pulsar.energy_integrated(t)
        volume = 4.0 / 3 * np.pi * self.radius(t) ** 3
        return np.sqrt(2 * const.mu0 * self.eta_B * energy / volume)

    def luminosity_tev(self, t=None, fraction=0.1):
        """TeV luminosity from a simple evolution model.

        Assumes that the luminosity is just a fraction of the total energy content
        of the pulsar. No cooling is considered and therefore the estimate is very bad.

        Parameters
        ----------
        t : `~astropy.units.Quantity`
            Time after birth of the SNR.
        """
        return fraction * self.pulsar.energy_integrated(t)



"""Test if Jupyter notebooks work."""
import os
import sys
import logging
from pkg_resources import working_set
import yaml
from ..extern.pathlib import Path
from ..scripts.jupyter import notebook_test

log = logging.getLogger(__name__)


def get_notebooks():
    """Read `notebooks.yaml` info."""
    path = Path("tutorials") / "notebooks.yaml"
    with path.open() as fh:
        return yaml.safe_load(fh)


def requirement_missing(notebook):
    """Check if one of the requirements is missing."""
    if "requires" in notebook:
        if notebook["requires"] is None:
            return False
        for package in notebook["requires"].split():
            try:
                working_set.require(package)
            except Exception:
                return True
    return False


def main():
    logging.basicConfig(level=logging.INFO)

    if "GAMMAPY_DATA" not in os.environ:
        log.info("GAMMAPY_DATA environment variable not set.")
        log.info("Running notebook tests requires this environment variable.")
        log.info("Exiting now.")
        sys.exit()

    passed = True

    for notebook in get_notebooks():
        if requirement_missing(notebook):
            log.info(
                "Skipping notebook {} because requirement is missing.".format(
                    notebook["name"]
                )
            )
            continue

        filename = notebook["name"] + ".ipynb"
        path = Path("tutorials") / filename

        if not notebook_test(path):
            passed = False

    assert passed


if __name__ == "__main__":
    main()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
from numpy.testing import assert_allclose
from .. import Parameter, Parameters, optimize_iminuit

pytest.importorskip("iminuit")


def fcn(parameters):
    x = parameters["x"].value
    y = parameters["y"].value
    z = parameters["z"].value
    x_opt, y_opt, z_opt = 2, 3e2, 4e-2
    return (x - x_opt) ** 2 + (y - y_opt) ** 2 + (z - z_opt) ** 2


@pytest.fixture()
def pars():
    x = Parameter("x", 2.1)
    y = Parameter("y", 3.1, scale=1e2)
    z = Parameter("z", 4.1, scale=1e-2)
    return Parameters([x, y, z])


def test_iminuit_basic(pars):
    factors, info, minuit = optimize_iminuit(function=fcn, parameters=pars)

    assert info["success"]
    assert_allclose(fcn(pars), 0, atol=1e-5)

    # Check the result in parameters is OK
    assert_allclose(pars["x"].value, 2, rtol=1e-3)
    assert_allclose(pars["y"].value, 3e2, rtol=1e-3)
    # Precision of estimate on "z" is very poor (0.040488). Why is it so bad?
    assert_allclose(pars["z"].value, 4e-2, rtol=2e-2)

    # Check that minuit sees the parameter factors correctly
    assert_allclose(factors, [2, 3, 4], rtol=1e-3)
    assert_allclose(minuit.values["par_000_x"], 2, rtol=1e-3)
    assert_allclose(minuit.values["par_001_y"], 3, rtol=1e-3)
    assert_allclose(minuit.values["par_002_z"], 4, rtol=1e-3)


def test_iminuit_frozen(pars):
    pars["y"].frozen = True

    factors, info, minuit = optimize_iminuit(function=fcn, parameters=pars)

    assert info["success"]

    assert_allclose(pars["y"].value, 3.1e2)
    assert minuit.list_of_fixed_param() == ["par_001_y"]


def test_iminuit_limits(pars):
    pars["y"].min = 301

    factors, info, minuit = optimize_iminuit(function=fcn, parameters=pars)

    assert info["success"]

    # Check the result in parameters is OK
    assert_allclose(pars["x"].value, 2, rtol=1e-2)
    assert_allclose(pars["y"].value, 301, rtol=1e-3)

    # Check that minuit sees the limit factors correctly
    states = minuit.get_param_states()
    assert not states[0]["has_limits"]

    y = states[1]
    assert y["has_limits"]
    assert_allclose(y["lower_limit"], 3.01)

    # The next assert can be added when we no longer test on iminuit 1.2
    # See https://github.com/gammapy/gammapy/pull/1771
    # assert states[1]["upper_limit"] is None



"""Compute some statistics directly using these methods:
sherpa.stats.Stat.calc_stat(self, data, model, staterror=None, syserror=None, weight=None)
sherpa.stats.Stat.calc_staterror(self, data)


data = np.array([1, 2])
model = np.array([3,4])
staterror = np.array([4,5])
stat.calc_stat(data, model, staterror)

# returns (25.0, 5.0)
#o this is what is going on:
def calc_stat(data, model, staterror):
    fvec = (model - data) / staterror
    stat = (fvec ** 2).sum()
    return stat, fvec
# This is not what is described at
# http://cxc.cfa.harvard.edu/sherpa/statistics/#userstat
There they write
   fvec = ((data - model) / staterror)**2
   stat = fvec.sum()
 which results in a different fvec, but the same stat.

"""
import sherpa.stats as ss

stats = [('cash', ss.Cash),
         ('cstat', ss.CStat),
         ('chi2constvar', ss.Chi2ConstVar),
         ('chi2datavar', ss.Chi2DataVar),
         ('chi2gehrels', ss.Chi2Gehrels),
         ('chi2modvar', ss.Chi2ModVar),
         ('chi2xspecvar', ss.Chi2XspecVar)]

for stat_name, stat_class in stats:
    stat_object = stat_class()
    try:
        stat_error = stat_object.calc_staterror(0)
    except ValueError:
        stat_error = -999
    print('%20s %20.10f' % (stat_name, stat_error))
    # stat_object.calc_staterror(data=2, model=12, staterror=2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from astropy.table import Table, Column
import astropy.units as u
from ..stats import excess_matching_significance_on_off
from .models import PowerLaw
from .utils import CountsPredictor

__all__ = ["SensitivityEstimator"]


class SensitivityEstimator(object):
    """Estimate differential sensitivity.

    Uses a 1D spectral analysis and on / off measurement.

    Parameters
    ----------
    arf : `~gammapy.irf.EffectiveAreaTable`
        1D effective area
    rmf : `~gammapy.irf.EnergyDispersion`
        energy dispersion table
    bkg : `~gammapy.spectrum.CountsSpectrum`
        the background array
    livetime : `~astropy.units.Quantity`
        Livetime (object with the units of time), e.g. 5*u.h
    index : float, optional
        Index of the spectral shape (Power-law), should be positive (>0)
    alpha : float, optional
        On/OFF normalisation
    sigma : float, optional
        Minimum significance
    gamma_min : float, optional
        Minimum number of gamma-rays
    bkg_sys : float, optional
        Fraction of Background systematics relative to the number of ON counts

    An example can be found in :gp-notebook:`cta_sensitivity` .

    Notes
    -----
    This class allows to determine for each reconstructed energy bin the flux associated to the number of gamma-ray
    events for which the significance is ``sigma``, and being larger than ``gamma_min`` and ``bkg_sys`` percent larger than the
    number of background events in the ON region.
    """

    def __init__(
        self,
        arf,
        rmf,
        bkg,
        livetime,
        index=2.0,
        alpha=0.2,
        sigma=5.0,
        gamma_min=10.0,
        bkg_sys=0.05,
    ):
        self.arf = arf
        self.rmf = rmf
        self.bkg = bkg
        self.livetime = u.Quantity(livetime).to("s")
        self.index = index
        self.alpha = alpha
        self.sigma = sigma
        self.gamma_min = gamma_min
        self.bkg_sys = bkg_sys

        self._results_table = None

    @property
    def results_table(self):
        """Results table (`~astropy.table.Table`)."""
        return self._results_table

    def run(self):
        """Run the computation."""

        # TODO: let the user decide on energy binning
        # then integrate bkg model and gamma over those energy bins.
        energy = self.rmf.e_reco.log_center()

        bkg_counts = (self.bkg.data.data.to("1/s") * self.livetime).value

        excess_counts = excess_matching_significance_on_off(
            n_off=bkg_counts / self.alpha, alpha=self.alpha, significance=self.sigma
        )
        is_gamma_limited = excess_counts < self.gamma_min
        excess_counts[is_gamma_limited] = self.gamma_min

        model = PowerLaw(
            index=self.index, amplitude="1 cm-2 s-1 TeV-1", reference="1 TeV"
        )

        # TODO: simplify the following computation
        predictor = CountsPredictor(
            model, aeff=self.arf, edisp=self.rmf, livetime=self.livetime
        )
        predictor.run()
        counts = predictor.npred.data.data.value
        phi_0 = excess_counts / counts

        dnde_model = model(energy=energy)
        diff_flux = (phi_0 * dnde_model * energy ** 2).to("erg / (cm2 s)")

        # TODO: take self.bkg_sys into account
        # and add a criterion 'bkg sys'
        criterion = []
        for idx in range(len(energy)):
            if is_gamma_limited[idx]:
                c = "gamma"
            else:
                c = "significance"
            criterion.append(c)

        table = Table(
            [
                Column(
                    data=energy,
                    name="energy",
                    format="5g",
                    description="Reconstructed Energy",
                ),
                Column(
                    data=diff_flux,
                    name="e2dnde",
                    format="5g",
                    description="Energy squared times differential flux",
                ),
                Column(
                    data=excess_counts,
                    name="excess",
                    format="5g",
                    description="Number of excess counts in the bin",
                ),
                Column(
                    data=bkg_counts,
                    name="background",
                    format="5g",
                    description="Number of background counts in the bin",
                ),
                Column(
                    data=criterion,
                    name="criterion",
                    description="Sensitivity-limiting criterion",
                ),
            ]
        )
        self._results_table = table
        return table



"""Example how to compute and plot reflected regions."""
import matplotlib.pyplot as plt
import numpy as np
from astropy.coordinates import SkyCoord, Angle
from regions import CircleSkyRegion
from gammapy.maps import WcsNDMap
from gammapy.background import ReflectedRegionsFinder

# Exclude a rectangular region
exclusion_mask = WcsNDMap.create(npix=(801, 701), binsz=0.01, skydir=(83.6, 23.0))

coords = exclusion_mask.geom.get_coord().skycoord
mask = (Angle("23d") < coords.dec) & (coords.dec < Angle("24d"))
exclusion_mask.data = np.invert(mask)

pos = SkyCoord(83.633, 22.014, unit="deg")
radius = Angle(0.3, "deg")
on_region = CircleSkyRegion(pos, radius)
center = SkyCoord(83.633, 24, unit="deg")

# One can impose a minimal distance between ON region and first reflected regions
finder = ReflectedRegionsFinder(
    region=on_region,
    center=center,
    exclusion_mask=exclusion_mask,
    min_distance_input="0.2 rad",
)
finder.run()

fig1 = plt.figure(1)
finder.plot(fig=fig1)

# One can impose a minimal distance between two adjacent regions
finder = ReflectedRegionsFinder(
    region=on_region,
    center=center,
    exclusion_mask=exclusion_mask,
    min_distance_input="0.2 rad",
    min_distance="0.1 rad",
)
finder.run()
fig2 = plt.figure(2)
finder.plot(fig=fig2)

# One can impose a maximal number of regions to be extracted
finder = ReflectedRegionsFinder(
    region=on_region,
    center=center,
    exclusion_mask=exclusion_mask,
    min_distance_input="0.2 rad",
    max_region_number=5,
    min_distance="0.1 rad",
)
finder.run()
fig3 = plt.figure(3)
finder.plot(fig=fig3)

plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""1D spectrum analysis."""
from .core import *
from .observation import *
from .cosmic_ray import *
from .crab import *
from .energy_group import *
from .flux_point import *
from .utils import *
from .extract import *
from .simulation import *
from .fit import *
from .results import *
from .sensitivity import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
from ...utils.testing import requires_data, requires_dependency
from ...utils.random import get_random_state
from ...irf import EffectiveAreaTable
from ...spectrum import (
    PHACountsSpectrum,
    SpectrumObservationList,
    SpectrumObservation,
    SpectrumFit,
    SpectrumFitResult,
    models,
)


@requires_dependency("sherpa")
class TestFit:
    """Test fit on counts spectra without any IRFs"""

    def setup(self):
        self.nbins = 30
        binning = np.logspace(-1, 1, self.nbins + 1) * u.TeV
        self.source_model = models.PowerLaw(
            index=2, amplitude=1e5 / u.TeV, reference=0.1 * u.TeV
        )
        self.bkg_model = models.PowerLaw(
            index=3, amplitude=1e4 / u.TeV, reference=0.1 * u.TeV
        )

        self.alpha = 0.1
        random_state = get_random_state(23)
        npred = self.source_model.integral(binning[:-1], binning[1:])
        source_counts = random_state.poisson(npred)
        self.src = PHACountsSpectrum(
            energy_lo=binning[:-1],
            energy_hi=binning[1:],
            data=source_counts,
            backscal=1,
        )
        # Currently it's necessary to specify a lifetime
        self.src.livetime = 1 * u.s

        npred_bkg = self.bkg_model.integral(binning[:-1], binning[1:])

        bkg_counts = random_state.poisson(npred_bkg)
        off_counts = random_state.poisson(npred_bkg * 1.0 / self.alpha)
        self.bkg = PHACountsSpectrum(
            energy_lo=binning[:-1], energy_hi=binning[1:], data=bkg_counts
        )
        self.off = PHACountsSpectrum(
            energy_lo=binning[:-1],
            energy_hi=binning[1:],
            data=off_counts,
            backscal=1.0 / self.alpha,
        )

    def test_cash(self):
        """Simple CASH fit to the on vector"""
        obs = SpectrumObservation(on_vector=self.src)
        obs_list = SpectrumObservationList([obs])

        fit = SpectrumFit(
            obs_list=obs_list,
            model=self.source_model,
            stat="cash",
            forward_folded=False,
        )
        assert "Spectrum" in str(fit)

        fit.predict_counts()
        assert_allclose(fit.predicted_counts[0][5], 660.5171, rtol=1e-5)

        fit.calc_statval()
        assert_allclose(np.sum(fit.statval[0]), -107346.5291, rtol=1e-5)

        self.source_model.parameters["index"].value = 1.12
        fit.run()
        # These values are check with sherpa fits, do not change
        pars = fit.result[0].model.parameters
        assert_allclose(pars["index"].value, 1.995525, rtol=1e-3)
        assert_allclose(pars["amplitude"].value, 100245.9, rtol=1e-3)

    def test_wstat(self):
        """WStat with on source and background spectrum"""
        on_vector = self.src.copy()
        on_vector.data.data += self.bkg.data.data
        obs = SpectrumObservation(on_vector=on_vector, off_vector=self.off)
        obs_list = SpectrumObservationList([obs])

        self.source_model.parameters.index = 1.12
        fit = SpectrumFit(
            obs_list=obs_list,
            model=self.source_model,
            stat="wstat",
            forward_folded=False,
        )
        fit.run()
        pars = fit.result[0].model.parameters
        assert_allclose(pars["index"].value, 1.997342, rtol=1e-3)
        assert_allclose(pars["amplitude"].value, 100245.187067, rtol=1e-3)
        assert_allclose(fit.result[0].statval, 30.022316, rtol=1e-3)

    def test_joint(self):
        """Test joint fit for obs with different energy binning"""
        obs1 = SpectrumObservation(on_vector=self.src)
        src_rebinned = self.src.rebin(2)
        obs2 = SpectrumObservation(on_vector=src_rebinned)
        fit = SpectrumFit(
            obs_list=[obs1, obs2],
            stat="cash",
            model=self.source_model,
            forward_folded=False,
        )
        fit.run()
        pars = fit.result[0].model.parameters
        assert_allclose(pars["index"].value, 1.996456, rtol=1e-3)

    def test_fit_range(self):
        """Test fit range without complication of thresholds"""
        obs = SpectrumObservation(on_vector=self.src)
        obs_list = SpectrumObservationList([obs])

        fit = SpectrumFit(
            obs_list=obs_list, model=self.source_model, stat=None, forward_folded=False
        )
        assert np.sum(fit._bins_in_fit_range[0]) == self.nbins
        assert_allclose(fit.true_fit_range[0][-1].value, 10)
        assert_allclose(fit.true_fit_range[0][0].value, 0.1)

        fit.fit_range = [200, 600] * u.GeV
        assert np.sum(fit._bins_in_fit_range[0]) == 6
        assert_allclose(fit.true_fit_range[0][0].value, 0.21544347, rtol=1e-5)
        assert_allclose(fit.true_fit_range[0][-1].value, 0.54117, rtol=1e-5)

        fit.fit_range = [0.11659144 + 1.0e-5, 1.0 - 1.0e-5] * u.TeV
        assert np.sum(fit._bins_in_fit_range[0]) == 14

        # Check different fit ranges for different observations
        on_vector2 = self.src.copy()
        obs2 = SpectrumObservation(on_vector=on_vector2)
        obs2.lo_threshold = 5 * u.TeV
        obs_list.append(obs2)
        fit = SpectrumFit(
            obs_list=obs_list, model=self.source_model, stat=None, forward_folded=False
        )
        fit.fit_range = [2, 8] * u.TeV
        assert_allclose(fit.true_fit_range[0][0].value, 2.15443, rtol=1e-3)
        assert_allclose(fit.true_fit_range[1][0].value, 5.41169, rtol=1e-3)

    def test_likelihood_profile(self):
        obs = SpectrumObservation(on_vector=self.src)
        fit = SpectrumFit(
            obs_list=obs, stat="cash", model=self.source_model, forward_folded=False
        )
        result = fit.run()
        true_idx = result.model.parameters["index"].value
        values = np.linspace(0.95 * true_idx, 1.05 * true_idx, 100)
        profile = fit.likelihood_profile("index", values=values)
        actual = values[np.argmin(profile["likelihood"])]
        assert_allclose(actual, true_idx, rtol=0.01)


@requires_dependency("sherpa")
@requires_data("gammapy-data")
class TestSpectralFit:
    """Test fit in astrophysical scenario"""

    def setup(self):
        path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
        obs1 = SpectrumObservation.read(path + "pha_obs23523.fits")
        obs2 = SpectrumObservation.read(path + "pha_obs23592.fits")
        self.obs_list = SpectrumObservationList([obs1, obs2])

        self.pwl = models.PowerLaw(
            index=2, amplitude=1e-12 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
        )

        self.ecpl = models.ExponentialCutoffPowerLaw(
            index=2,
            amplitude=1e-12 * u.Unit("cm-2 s-1 TeV-1"),
            reference=1 * u.TeV,
            lambda_=0.1 / u.TeV,
        )

        # Example fit for one observation
        self.fit = SpectrumFit(self.obs_list[0], self.pwl)

    @requires_dependency("iminuit")
    def test_basic_results(self):
        self.fit.run()
        result = self.fit.result[0]
        pars = result.model.parameters

        assert self.pwl is self.fit._model
        assert self.fit._model is result.model
        assert self.pwl is result.model

        assert_allclose(result.statval, 38.343, rtol=1e-3)
        assert_allclose(pars["index"].value, 2.817, rtol=1e-3)
        assert pars["amplitude"].unit == "cm-2 s-1 TeV-1"
        assert_allclose(pars["amplitude"].value, 5.142e-11, rtol=1e-3)
        assert_allclose(result.npred[60], 0.6102, rtol=1e-3)
        self.fit.result[0].to_table()

    def test_basic_errors(self):
        self.fit.run()
        result = self.fit.result[0]
        pars = result.model.parameters
        assert_allclose(pars.error("index"), 0.1496, rtol=1e-3)
        assert_allclose(pars.error("amplitude"), 6.423e-12, rtol=1e-3)
        self.fit.result[0].to_table()

    def test_compound(self):
        model = self.pwl * 2
        fit = SpectrumFit(self.obs_list[0], model)
        fit.run()
        result = fit.result[0]
        pars = result.model.parameters
        assert_allclose(pars["index"].value, 2.8166, rtol=1e-3)
        p = pars["amplitude"]
        assert p.unit == "cm-2 s-1 TeV-1"
        assert_allclose(p.value, 5.0714e-12, rtol=1e-3)

    def test_npred(self):
        self.fit.run()
        actual = (
            self.fit.obs_list[0]
            .predicted_counts(self.fit.result[0].model)
            .data.data.value
        )
        desired = self.fit.result[0].npred
        assert_allclose(actual, desired)

    def test_stats(self):
        self.fit.run()
        stats = self.fit.result[0].stat_per_bin
        actual = np.sum(stats)
        desired = self.fit.result[0].statval
        assert_allclose(actual, desired)

    def test_fit_range(self):
        # Fit range not restriced fit range should be the thresholds
        obs = self.fit.obs_list[0]
        desired = obs.on_vector.lo_threshold
        actual = self.fit.true_fit_range[0][0]
        assert actual.unit == "keV"
        assert_allclose(actual.value, desired.value)

        # Restrict fit range
        fit_range = [4, 20] * u.TeV
        self.fit.fit_range = fit_range

        range_bin = obs.on_vector.energy.find_node(fit_range[1])
        desired = obs.on_vector.energy.lo[range_bin]
        actual = self.fit.true_fit_range[0][1]
        assert_allclose(actual.value, desired.value)

        # Make sure fit range is not extended below threshold
        fit_range = [0.001, 10] * u.TeV
        self.fit.fit_range = fit_range
        desired = obs.on_vector.lo_threshold
        actual = self.fit.true_fit_range[0][0]
        assert_allclose(actual.value, desired.value)

    def test_no_edisp(self):
        obs = self.obs_list[0]
        # Bring aeff in RECO space
        data = obs.aeff.data.evaluate(energy=obs.on_vector.energy.nodes)
        obs.aeff = EffectiveAreaTable(
            data=data,
            energy_lo=obs.on_vector.energy.lo,
            energy_hi=obs.on_vector.energy.hi,
        )
        obs.edisp = None
        fit = SpectrumFit(obs_list=obs, model=self.pwl)
        fit.run()
        assert_allclose(
            fit.result[0].model.parameters["index"].value, 2.7961, atol=0.02
        )

    def test_ecpl_fit(self):
        fit = SpectrumFit(self.obs_list[0], self.ecpl)
        fit.run()
        actual = fit.result[0].model.parameters["lambda_"].quantity
        assert actual.unit == "TeV-1"
        assert_allclose(actual.value, 0.145215, rtol=1e-2)

    def test_joint_fit(self):
        fit = SpectrumFit(self.obs_list, self.pwl)
        fit.run()
        actual = fit.result[0].model.parameters["index"].value
        assert_allclose(actual, 2.7806, rtol=1e-3)

        actual = fit.result[0].model.parameters["amplitude"].quantity
        assert actual.unit == "cm-2 s-1 TeV-1"
        assert_allclose(actual.value, 5.200e-11, rtol=1e-3)

    def test_stacked_fit(self):
        stacked_obs = self.obs_list.stack()
        obs_list = SpectrumObservationList([stacked_obs])
        fit = SpectrumFit(obs_list, self.pwl)
        fit.run()
        pars = fit.result[0].model.parameters
        assert_allclose(pars["index"].value, 2.7767, rtol=1e-3)
        assert u.Unit(pars["amplitude"].unit) == "cm-2 s-1 TeV-1"
        assert_allclose(pars["amplitude"].value, 5.191e-11, rtol=1e-3)

    def test_run(self, tmpdir):
        fit = SpectrumFit(self.obs_list, self.pwl)
        fit.run()

        result = fit.result[0]
        modelname = result.model.__class__.__name__
        filename = tmpdir / "fit_result_{}.yaml".format(modelname)
        result.to_yaml(filename)

        read_result = SpectrumFitResult.from_yaml(tmpdir / "fit_result_PowerLaw.yaml")

        desired = fit.result[0].model.evaluate_error(1 * u.TeV)
        actual = read_result.model.evaluate_error(1 * u.TeV)
        assert actual.unit == desired.unit
        assert_allclose(actual.value, desired.value)

    def test_sherpa_fit(self, tmpdir):
        # this is to make sure that the written PHA files work with sherpa
        import sherpa.astro.ui as sau
        from sherpa.models import PowLaw1D

        # TODO: this works a little bit, but some info and warnings
        # from Sherpa remain. Not sure what to do, OK as-is for now.
        import logging

        logging.getLogger("sherpa").setLevel("ERROR")

        self.obs_list.write(tmpdir, use_sherpa=True)
        filename = tmpdir / "pha_obs23523.fits"
        sau.load_pha(str(filename))
        sau.set_stat("wstat")
        model = PowLaw1D("powlaw1d.default")
        model.ref = 1e9
        model.ampl = 1
        model.gamma = 2
        sau.set_model(model * 1e-20)
        sau.fit()
        assert_allclose(model.pars[0].val, 2.732, rtol=1e-3)
        assert_allclose(model.pars[2].val, 4.647, rtol=1e-3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from astropy import units as u
from astropy.modeling.models import Gaussian2D
from astropy.coordinates import SkyCoord
from ...utils.testing import assert_quantity_allclose
from ...maps import WcsGeom, WcsNDMap
from ...image import (
    measure_containment_radius,
    measure_image_moments,
    measure_containment,
    measure_curve_of_growth,
)


@pytest.fixture(scope="session")
def gaussian_image():
    """Generate gaussian test image.
    """
    binsz = 0.02
    sigma = 0.2
    geom = WcsGeom.create(npix=(201, 201), binsz=binsz, coordsys="GAL")
    gauss = Gaussian2D(1.0 / (2 * np.pi * (sigma / binsz) ** 2), 0, 0, sigma, sigma)
    coord = geom.get_coord().skycoord
    l = coord.data.lon.wrap_at("180d")
    b = coord.data.lat
    data = gauss(l.degree, b.degree)
    return WcsNDMap(geom=geom, data=data, unit="cm-2 s-1")


def test_measure_image_moments(gaussian_image):
    """Test measure_image_moments function"""
    moments = measure_image_moments(gaussian_image)

    reference = [
        1 * u.Unit("cm-2 s-1"),
        0 * u.deg,
        0 * u.deg,
        0.2 * u.deg,
        0.2 * u.deg,
        0.2 * u.deg,
    ]

    for val, ref in zip(moments, reference):
        assert_quantity_allclose(val, ref, atol=1e-12 * val.unit)


def test_measure_containment(gaussian_image):
    """Test measure_containment function"""
    position = SkyCoord(0, 0, frame="galactic", unit="deg")
    radius = u.Quantity(0.2 * np.sqrt(2 * np.log(5)), "deg")
    frac = measure_containment(gaussian_image, position, radius)
    ref = u.Quantity(0.8, "cm-2 s-1")
    assert_quantity_allclose(frac, ref, rtol=0.01)


def test_measure_containment_radius(gaussian_image):
    """Test measure_containment_radius function"""
    position = SkyCoord(0, 0, frame="galactic", unit="deg")
    rad = measure_containment_radius(gaussian_image, position, 0.8)
    ref = 0.2 * np.sqrt(2 * np.log(5)) * u.deg
    assert_quantity_allclose(rad, ref, rtol=0.01)


def test_measure_curve_of_growth(gaussian_image):
    """Test measure_curve_of_growth function"""
    position = SkyCoord(0, 0, frame="galactic", unit="deg")
    radius_max = 0.6 * u.deg
    radius, containment = measure_curve_of_growth(gaussian_image, position, radius_max)
    sigma = 0.2 * u.deg
    containment_ana = u.Quantity(
        1 - np.exp(-0.5 * (radius / sigma) ** 2).value, "cm-2 s-1"
    )
    assert_quantity_allclose(containment, containment_ana, rtol=0.1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This package contains `extern` code, i.e. code that we just copied
here into the `gammapy.extern` package, because we wanted to use it,
but not have an extra dependency (these are single-file external packages).

Alphabetical order:

* ``appdirs.py`` is a copy from Github on 2016-01-08
  Origin: https://raw.githubusercontent.com/ActiveState/appdirs/master/appdirs.py
  See also: https://pypi.org/project/appdirs/
* ``pathlib.py`` is a copy pathlib2 from github on 2015-10-20
  Origin https://raw.githubusercontent.com/mcmtroffaes/pathlib2/develop/pathlib2.py
  One line was patched: `import six` -> `from . import six`
  See also: https://pypi.org/project/pathlib2/
* ``six.py`` - Python 2/3 compatibility
* ``validator.py`` for `~astropy.units.Quantity` validation helper functions.
  Origin: https://github.com/astrofrog/sedfitter/blob/master/sedfitter/utils/validator.py
* ``xmltodict.py`` for easily converting XML from / to Python dicts
  Origin: https://github.com/martinblech/xmltodict/blob/master/xmltodict.py
* ``zeros.py`` - Modified copy of the file from ``scipy.optimize``,
  used by the TS map computation code.
* ``skimage.py`` - utility functions copied from scikit-image
  They have the same BSD license as we do.
"""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from .obs_stats import ObservationStats

__all__ = ["ObservationTableSummary", "ObservationSummary"]


class ObservationTableSummary(object):
    """Observation table summary.

    Class allowing to summarize informations contained in
    Observation index table (`~gammapy.data.ObservationTable`)

    Parameters
    ----------
    obs_table : `~gammapy.data.ObservationTable`
        Observation index table
    target_pos : `~astropy.coordinates.SkyCoord`
        Target position
    """

    def __init__(self, obs_table, target_pos=None):
        self.obs_table = obs_table
        self.target_pos = target_pos

    @property
    def offset(self):
        """Observation pointing ot target offset (`~astropy.coordinates.Angle`).
        """
        t = self.obs_table
        pnt_pos = SkyCoord(t["RA_PNT"], t["DEC_PNT"], unit="deg")

        return pnt_pos.separation(self.target_pos)

    def __str__(self):
        ss = "*** Observation summary ***\n"
        ss += "Target position: {}\n".format(self.target_pos)

        ss += "Number of observations: {}\n".format(len(self.obs_table))

        livetime = u.Quantity(sum(self.obs_table["LIVETIME"]), "second")
        ss += "Livetime: {:.2f}\n".format(livetime.to("hour"))
        zenith = self.obs_table["ZEN_PNT"]
        ss += "Zenith angle: (mean={:.2f}, std={:.2f})\n".format(
            zenith.mean(), zenith.std()
        )
        offset = self.offset
        ss += "Offset: (mean={:.2f}, std={:.2f})\n".format(offset.mean(), offset.std())

        return ss

    def plot_zenith_distribution(self, ax=None, bins=None):
        """Construct the zenith distribution of the observations.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.
            If None, uses the current `~matplotlib.axes.Axes`.
        bins : integer or array_like, optional
            Binning specification, passed to `matplotlib.pyplot.hist`.
            By default, 30 bins from 0 deg to max zenith + 5 deg is used.

        Returns
        --------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax

        zenith = self.obs_table["ZEN_PNT"]

        if bins is None:
            bins = np.linspace(0, zenith.max() + 5, 30)

        ax.hist(zenith, bins=bins)
        ax.set_title("Zenith distribution")
        ax.set_xlabel("Zenith (Deg)")
        ax.set_ylabel("#Entries")

        return ax

    def plot_offset_distribution(self, ax=None, bins=None):
        """Construct the offset distribution of the observations.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.
            If None, uses the current `~matplotlib.axes.Axes`.
        bins : integer or array_like, optional
            Binning specification, passed to `matplotlib.pyplot.hist`.
            By default, 10 bins from 0 deg to max offset + 0.5 deg is used.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax

        offset = self.offset

        if bins is None:
            bins = np.linspace(0, offset.degree.max() + 0.5, 10)

        ax.hist(offset.degree, bins=bins)
        ax.set_title("Offset distribution")
        ax.set_xlabel("Offset (Deg)")
        ax.set_ylabel("#Entries")

        return ax


class ObservationSummary(object):
    """Summary of observations.

    For a list of observation stats, this class can make a
    table and do summary printout and plots.

    * TODO: Data should be stored in `~astropy.table.Table`!
    * TODO: there should be a per-run version of the stats
      in addition to the cumulative version that's there now.

    Parameters
    ----------
    obs_stats : list
        Python list of `~gammapy.data.ObservationStats`
    """

    def __init__(self, obs_stats):
        self.obs_stats = obs_stats

        self.obs_id = np.zeros(len(self.obs_stats))
        self.livetime = np.zeros(len(self.obs_stats)) * u.s
        self.n_on = np.zeros(len(self.obs_stats))
        self.n_off = np.zeros(len(self.obs_stats))
        self.alpha = np.zeros(len(self.obs_stats))
        self.background = np.zeros(len(self.obs_stats))
        self.excess = np.zeros(len(self.obs_stats))
        self.sigma = np.zeros(len(self.obs_stats))
        self.gamma_rate = np.zeros(len(self.obs_stats)) / u.min
        self.bg_rate = np.zeros(len(self.obs_stats)) / u.min

        self._init_values()

    def _init_values(self):
        """Initialise vector attributes for plotting methods."""
        stats_list = []

        for index, obs in enumerate(self.obs_stats):
            # per observation stat
            self.obs_id[index] = obs.obs_id  # keep track of the observation
            self.gamma_rate[index] = obs.gamma_rate
            self.bg_rate[index] = obs.bg_rate

            # cumulative information
            stats_list.append(obs)
            stack = ObservationStats.stack(stats_list)
            self.livetime[index] = stack.livetime
            self.n_on[index] = stack.n_on
            self.n_off[index] = stack.n_off
            self.alpha[index] = stack.alpha
            self.background[index] = stack.background
            self.excess[index] = stack.excess
            self.sigma[index] = stack.sigma

    def obs_wise_summary(self):
        """Observation wise summary report (str)."""
        ss = "*** Observation Wise summary ***\n"
        for obs in self.obs_stats:
            ss += "{}\n".format(obs)

        return ss

    def __str__(self):
        stack = ObservationStats.stack(self.obs_stats)
        ss = "*** Observation summary ***\n"
        ss += "{}\n".format(stack)
        return ss

    def plot_significance_vs_livetime(self, ax=None, **kwargs):
        """Plot significance as a function of livetime.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax
        ax.plot(self.livetime.to(u.h), self.sigma, "o", **kwargs)

        ax.set_xlabel("Livetime ({})".format(u.h))
        ax.set_ylabel("Significance")

        xmax = np.amax(self.livetime.to_value("h")) * 1.2
        ymax = np.amax(self.sigma) * 1.2
        ax.axis([0, xmax, 0, ymax])
        ax.set_title("Significance evolution")
        return ax

    def plot_excess_vs_livetime(self, ax=None, **kwargs):
        """Plot excess as a function of livetime.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax
        ax.plot(self.livetime.to(u.h), self.excess, "o", **kwargs)

        ax.set_xlabel("Livetime ({})".format(u.h))
        ax.set_ylabel("Excess")

        xmax = np.amax(self.livetime.to_value("h")) * 1.2
        ymax = np.amax(self.excess) * 1.2
        ax.axis([0, xmax, 0, ymax])
        ax.set_title("Excess evolution")
        return ax

    def plot_background_vs_livetime(self, ax=None, **kwargs):
        """Plot background as a function of livetime.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax
        ax.plot(self.livetime.to(u.h), self.background, "o", **kwargs)

        ax.set_xlabel("Livetime ({})".format(u.h))
        ax.set_ylabel("Background")

        xmax = np.amax(self.livetime.to_value("h")) * 1.2
        ymax = np.amax(self.background) * 1.2
        ax.axis([0, xmax, 0, ymax])
        ax.set_title("Background evolution")
        return ax

    def plot_gamma_rate(self, ax=None, **kwargs):
        """Plot gamma rate for each observation.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax

        xtick_vals, xtick_labels = self._get_xtick_info()

        ax.plot(xtick_vals, self.gamma_rate, "o", **kwargs)
        ax.set_xlabel("Observation ID")

        ax.set_xticks(xtick_vals)
        ax.set_xticklabels(xtick_labels, rotation=-22.5)
        ax.set_ylabel("Excess rate ({})".format(self.gamma_rate.unit))
        ax.axis([0, len(self.gamma_rate), 0.0, np.amax(self.gamma_rate.value) * 1.2])
        ax.set_title("Excess rates")
        return ax

    def plot_background_rate(self, ax=None, **kwargs):
        """Plot background rate for each observation.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` or None, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax

        xtick_vals, xtick_labels = self._get_xtick_info()

        ax.plot(xtick_vals, self.bg_rate, "o", **kwargs)
        ax.set_xlabel("Observation Ids")

        ax.set_xticks(xtick_vals)
        ax.set_xticklabels(xtick_labels, rotation=-22.5)
        ax.set_ylabel("Background rate ({})".format(self.bg_rate.unit))
        ax.axis([0, len(self.bg_rate), 0.0, np.amax(self.bg_rate.value) * 1.2])
        ax.set_title("Background rates")
        return ax

    def _get_xtick_info(self):
        idxs = list(range(len(self.obs_stats)))
        vals = [idx + 0.5 for idx in idxs]
        labels = [str(int(self.obs_id[idx])) for idx in idxs]
        return vals, labels



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from regions import CircleSkyRegion
from .ring import ring_area_factor

__all__ = ["BackgroundEstimate", "ring_background_estimate"]


class BackgroundEstimate(object):
    """Container class for background estimate.

    This container holds the result from a region based background estimation
    for one observation. Currently, it is filled by the functions
    :func:`~gammapy.background.ring_background_estimate` and
    the `~gammapy.background.ReflectedRegionsBackgroundEstimator`.

    Parameters
    ----------
    on_region : `~regions.SkyRegion`
        Signal extraction region
    on_events : `~gammapy.data.EventList`
        Signal events
    off_region : `~regions.SkyRegion`
        Background extraction region
    off_events : `~gammapy.data.EventList`
        Background events
    a_on : float
        Relative background exposure of the on region
    a_off : float
        Relative background exposure of the off region
    method : str
        Background estimation method
    """

    def __init__(
        self,
        on_region,
        on_events,
        off_region,
        off_events,
        a_on,
        a_off,
        method="default",
    ):
        self.on_region = on_region
        self.on_events = on_events
        self.off_region = off_region
        self.off_events = off_events
        self.a_on = a_on
        self.a_off = a_off
        self.method = method

    def __str__(self):
        ss = self.__class__.__name__
        ss += "\n Method: {}".format(self.method)
        ss += "\n on region"
        ss += "\n {}".format(self.on_region)
        ss += "\n {}".format(self.on_events)
        ss += "\n off region"
        ss += "\n {}".format(self.off_region)
        ss += "\n {}".format(self.off_events)
        return ss


def ring_background_estimate(pos, on_radius, inner_radius, outer_radius, events):
    """Simple ring background estimate.

    No acceptance correction is applied

    TODO : Replace with AnnulusSkyRegion

    Parameters
    ----------
    pos : `~astropy.coordinates.SkyCoord`
        On region radius
    on_radius : `~astropy.coordinates.Angle`
        On region radius
    inner_radius, outer_radius : `~astropy.coordinates.Angle`
        Inner and outer ring radius
    events : `~gammapy.data.EventList`
        Event list

    Returns
    -------
    bkg : `~gammapy.data.BackgroundEstimate`
        Background estimate
    """
    on_region = CircleSkyRegion(center=pos, radius=on_radius)
    on_events = events.select_circular_region(on_region)

    off_region = dict(inner=inner_radius, outer=outer_radius)
    off_events = events.select_sky_ring(pos, inner_radius, outer_radius)

    # TODO: change to region areas here (e.g. in steratian?)
    a_on = 1
    a_off = ring_area_factor(on_radius, inner_radius, outer_radius).value

    return BackgroundEstimate(
        on_region=on_region,
        on_events=on_events,
        off_region=off_region,
        off_events=off_events,
        a_on=a_on,
        a_off=a_off,
        method="ring",
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from ..extern import six
from astropy.utils.data import download_file
from astropy.coordinates import Angle, SkyCoord
from astropy.table import Table, Column
from .core import SourceCatalog, SourceCatalogObject

__all__ = ["SourceCatalogSNRcat", "SourceCatalogObjectSNRcat"]


class SourceCatalogObjectSNRcat(SourceCatalogObject):
    """One source from the SNRcat catalog."""

    pass


class SourceCatalogSNRcat(SourceCatalog):
    """SNRcat supernova remnant catalog.

    `SNRcat <http://www.physics.umanitoba.ca/snr/SNRcat/>`__
    is a census of high-energy observations of Galactic supernova remnants.

    This function downloads the following CSV-format tables
    and adds some useful columns and unit information:

    * http://www.physics.umanitoba.ca/snr/SNRcat/SNRdownload.php?table=SNR
    * http://www.physics.umanitoba.ca/snr/SNRcat/SNRdownload.php?table=OBS

    This only represents a subset of the information available in SNRcat,
    to get at the rest we would have to scrape their web pages.

    * ``table`` (`~astropy.table.Table`) -- SNR info table
    * ``obs_table`` (`~astropy.table.Table`) -- High-energy observation info table

    Each table has a ``version`` string containing the download date in the ``table.meta`` dictionary.
    """

    name = "snrcat"
    description = "SNRcat supernova remnant catalog."
    source_object_class = SourceCatalogObjectSNRcat

    def __init__(self, cache=False):
        # TODO: load from gammapy-extra?
        # At least optionally?
        self.snr_table = _fetch_catalog_snrcat_snr_table(cache=cache)
        self.obs_table = _fetch_catalog_snrcat_obs_table(cache=cache)

        super(SourceCatalogSNRcat, self).__init__(table=self.snr_table)


def _fetch_catalog_snrcat_snr_table(cache):
    url = "http://www.physics.umanitoba.ca/snr/SNRcat/SNRdownload.php?table=SNR"
    filename = download_file(url, cache=cache)

    # Note: currently the first line contains this comment, which we skip via `header_start=1`
    table = Table.read(filename, format="ascii.csv", header_start=1, delimiter=";")
    table.meta["url"] = url
    table.meta["version"] = _snrcat_parse_download_date(filename)

    # TODO: doesn't work properly ... don't call for now.
    # _snrcat_fix_na(table)

    table.rename_column("G", "Source_Name")

    table.rename_column("J2000_ra (hh:mm:ss)", "RAJ2000_str")
    table.rename_column("J2000_dec (dd:mm:ss)", "DEJ2000_str")

    data = Angle(table["RAJ2000_str"], unit="hour").deg
    index = table.index_column("RAJ2000_str") + 1
    table.add_column(Column(data=data, name="RAJ2000", unit="deg"), index=index)

    data = Angle(table["DEJ2000_str"], unit="deg").deg
    index = table.index_column("DEJ2000_str") + 1
    table.add_column(Column(data=data, name="DEJ2000", unit="deg"), index=index)

    radec = SkyCoord(table["RAJ2000"], table["DEJ2000"], unit="deg")
    galactic = radec.galactic
    table.add_column(Column(data=galactic.l.deg, name="GLON", unit="deg"))
    table.add_column(Column(data=galactic.b.deg, name="GLAT", unit="deg"))

    table.rename_column("age_min (yr)", "age_min")
    table["age_min"].unit = "year"
    table.rename_column("age_max (yr)", "age_max")
    table["age_max"].unit = "year"
    distance = np.mean([table["age_min"], table["age_max"]], axis=0)
    index = table.index_column("age_max") + 1
    table.add_column(Column(distance, name="age", unit="year"), index=index)

    table.rename_column("distance_min (kpc)", "distance_min")
    table["distance_min"].unit = "kpc"
    table.rename_column("distance_max (kpc)", "distance_max")
    table["distance_max"].unit = "kpc"
    distance = np.mean([table["distance_min"], table["distance_max"]], axis=0)
    index = table.index_column("distance_max") + 1
    table.add_column(Column(distance, name="distance", unit="kpc"), index=index)

    table.rename_column("size_radio", "diameter_radio_str")
    diameter_radio_mean = _snrcat_parse_diameter(table["diameter_radio_str"])
    index = table.index_column("diameter_radio_str") + 1
    table.add_column(
        Column(diameter_radio_mean, name="diameter_radio_mean", unit="arcmin"),
        index=index,
    )

    table.rename_column("size_X", "diameter_xray_str")
    diameter_xray_mean = _snrcat_parse_diameter(table["diameter_xray_str"])
    index = table.index_column("diameter_xray_str") + 1
    table.add_column(
        Column(diameter_xray_mean, name="diameter_xray_mean", unit="arcmin"),
        index=index,
    )

    table.rename_column("size_coarse (arcmin)", "diameter_mean")
    table["diameter_mean"].unit = "arcmin"

    table.rename_column("size_imprecise", "diameter_mean_is_imprecise")

    return table


def _fetch_catalog_snrcat_obs_table(cache):
    url = "http://www.physics.umanitoba.ca/snr/SNRcat/SNRdownload.php?table=OBS"
    filename = download_file(url, cache=cache)

    # Note: currently the first line contains this comment, which we skip via `header_start=1`
    table = Table.read(filename, format="ascii.csv", header_start=1, delimiter=";")
    table.meta["url"] = url
    table.meta["version"] = _snrcat_parse_download_date(filename)

    # TODO: doesn't work properly ... don't call for now.
    # _snrcat_fix_na(table)

    return table


def _snrcat_fix_na(table):
    """Fix N/A entries in string columns in SNRcat."""
    for colname in table.colnames:
        if isinstance(table[colname][0], six.text_type):
            mask1 = table[colname] == "N / A"
            mask2 = table[colname] == "N/A"
            table[colname].mask = mask1 | mask2
            table[colname].fill_value = ""


def _snrcat_parse_download_date(filename):
    text = open(filename).readline()
    # Format: "This file was downloaded on 2015-06-07T03:39:53 CDT ..."
    tokens = text.split()
    date = tokens[5]
    return date[:10]


def _snrcat_parse_diameter(text_col):
    """Parse SNRcat diameter string column"""
    d_means = []
    for text in text_col:
        try:
            # Parse this text field:
            if "x" in text:
                a, b = text.split("x")
                d_major = Angle(a).arcmin
                d_minor = Angle(b).arcmin
            else:
                d_major = Angle(text).arcmin
                d_minor = d_major
            d_mean = _snr_mean_diameter(d_major, d_minor)
        except:
            # print('Parsing error:', text)
            d_mean = np.nan

        d_means.append(d_mean)

    return d_means


def _snr_mean_diameter(d_major, d_minor):
    """Compute geometric mean diameter (preserves area)"""
    diameter = np.sqrt(d_major * d_minor)
    # If no `d_minor` is given, use `d_major` as mean radius
    with np.errstate(invalid="ignore"):
        diameter = np.where(d_minor > 0, diameter, d_major)

    return diameter



# Licensed under a 3-clause BSD style license - see LICENSE.rst


def get_package_data():
    formats = "fits root xml json conf txt csv".split()
    files = ["data/*.{}".format(_) for _ in formats]
    return {"gammapy.irf.tests": files}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from numpy.testing import assert_allclose, assert_equal
import astropy.units as u
from ..core import PHACountsSpectrum
from ..observation import SpectrumObservation
from ..energy_group import (
    SpectrumEnergyGroups,
    SpectrumEnergyGroupMaker,
    SpectrumEnergyGroup,
)


class TestSpectrumEnergyGroup:
    @pytest.fixture()
    def group(self):
        return SpectrumEnergyGroup(3, 10, 20, "normal", 100 * u.TeV, 200 * u.TeV)

    def test_init(self, group):
        """Establish argument order in `__init__` and attributes."""
        assert group.energy_group_idx == 3
        assert group.bin_idx_min == 10
        assert group.bin_idx_max == 20
        assert group.bin_type == "normal"
        assert group.energy_min == 100 * u.TeV
        assert group.energy_max == 200 * u.TeV

    def test_repr(self, group):
        assert "SpectrumEnergyGroup" in repr(group)

    def test_to_dict(self, group):
        # Check that it round-trips
        group2 = SpectrumEnergyGroup(**group.to_dict())
        assert group2 == group

    def test_bin_idx_array(self, group):
        assert_equal(group.bin_idx_array, np.arange(10, 21))

    def test_bin_table(self, group):
        table = group.bin_table
        assert_equal(table["bin_idx"], np.arange(10, 21))
        assert_equal(table["energy_group_idx"], 3)
        assert_equal(table["bin_type"], "normal")


class TestSpectrumEnergyGroups:
    @pytest.fixture()
    def groups(self):
        return SpectrumEnergyGroups(
            [
                # energy_group_idx, bin_idx_min, bin_idx_max, bin_type, energy_min, energy_max
                SpectrumEnergyGroup(0, 10, 20, "normal", 100 * u.TeV, 210 * u.TeV),
                SpectrumEnergyGroup(1, 21, 25, "normal", 210 * u.TeV, 260 * u.TeV),
                SpectrumEnergyGroup(5, 26, 26, "normal", 260 * u.TeV, 270 * u.TeV),
                SpectrumEnergyGroup(6, 27, 30, "normal", 270 * u.TeV, 300 * u.TeV),
            ]
        )

    def test_repr(self, groups):
        assert repr(groups) == "SpectrumEnergyGroups(len=4)"

    def test_str(self, groups):
        txt = str(groups)
        assert "SpectrumEnergyGroups" in txt
        assert "energy_group_idx" in txt

    def test_copy(self, groups):
        """Make sure groups.copy() is a deep copy"""
        groups2 = groups.copy()
        groups2[0].bin_type = "spam"
        assert groups[0].bin_type == "normal"

    def test_group_table(self, groups):
        """Check that info to and from group table round-trips"""
        table = groups.to_group_table()
        groups2 = SpectrumEnergyGroups.from_group_table(table)
        assert groups2 == groups

    def test_from_total_table(self, groups):
        table = groups.to_total_table()
        groups2 = SpectrumEnergyGroups.from_total_table(table)
        assert groups2 == groups

    def test_energy_range(self, groups):
        actual = groups.energy_range
        expected = [100, 300] * u.TeV
        assert_allclose(actual, expected)
        assert actual.unit == "TeV"

    def test_energy_bounds(self, groups):
        actual = groups.energy_bounds
        expected = [100, 210, 260, 270, 300] * u.TeV
        assert_allclose(actual, expected)
        assert actual.unit == "TeV"


class TestSpectrumEnergyGroupMaker:
    @pytest.fixture(scope="session")
    def obs(self):
        """An example SpectrumObservation object for tests."""
        pha_ebounds = np.arange(1, 11) * u.TeV
        on_vector = PHACountsSpectrum(
            energy_lo=pha_ebounds[:-1],
            energy_hi=pha_ebounds[1:],
            data=np.zeros(len(pha_ebounds) - 1),
            livetime=99 * u.s,
        )
        return SpectrumObservation(on_vector=on_vector)

    def test_groups_from_obs(self, obs):
        seg = SpectrumEnergyGroupMaker(obs=obs)
        seg.groups_from_obs()
        groups = seg.groups

        assert len(groups) == obs.e_reco.nbins

    def test_compute_groups_fixed_basic(self, obs):
        ebounds = [1, 2, 10] * u.TeV
        seg = SpectrumEnergyGroupMaker(obs=obs)
        seg.compute_groups_fixed(ebounds=ebounds)
        groups = seg.groups

        expected = SpectrumEnergyGroups(
            [
                SpectrumEnergyGroup(0, 0, 0, "normal", 1 * u.TeV, 2 * u.TeV),
                SpectrumEnergyGroup(1, 1, 8, "normal", 2 * u.TeV, 10 * u.TeV),
            ]
        )

        assert groups == expected

    @pytest.mark.parametrize(
        "ebounds",
        [[1.8, 4.8, 7.2] * u.TeV, [2, 5, 7] * u.TeV, [2000, 5000, 7000] * u.GeV],
    )
    def test_compute_groups_fixed_edges(self, obs, ebounds):
        seg = SpectrumEnergyGroupMaker(obs=obs)
        seg.compute_groups_fixed(ebounds=ebounds)
        groups = seg.groups

        expected = SpectrumEnergyGroups(
            [
                SpectrumEnergyGroup(0, 0, 0, "underflow", 1 * u.TeV, 2 * u.TeV),
                SpectrumEnergyGroup(1, 1, 3, "normal", 2 * u.TeV, 5 * u.TeV),
                SpectrumEnergyGroup(2, 4, 5, "normal", 5 * u.TeV, 7 * u.TeV),
                SpectrumEnergyGroup(3, 6, 8, "overflow", 7 * u.TeV, 10 * u.TeV),
            ]
        )

        assert groups == expected

    def test_compute_groups_fixed_below_range(self, obs):
        ebounds = [0.7, 0.8, 1, 4] * u.TeV
        seg = SpectrumEnergyGroupMaker(obs=obs)
        seg.compute_groups_fixed(ebounds=ebounds)
        groups = seg.groups

        expected = SpectrumEnergyGroups(
            [
                SpectrumEnergyGroup(0, 0, 2, "normal", 1 * u.TeV, 4 * u.TeV),
                SpectrumEnergyGroup(1, 3, 8, "overflow", 4 * u.TeV, 10 * u.TeV),
            ]
        )

        assert groups == expected

    def test_compute_groups_fixed_above_range(self, obs):
        ebounds = [5, 7, 11, 13] * u.TeV
        seg = SpectrumEnergyGroupMaker(obs=obs)
        seg.compute_groups_fixed(ebounds=ebounds)
        groups = seg.groups

        expected = SpectrumEnergyGroups(
            [
                SpectrumEnergyGroup(0, 0, 3, "underflow", 1 * u.TeV, 5 * u.TeV),
                SpectrumEnergyGroup(1, 4, 5, "normal", 5 * u.TeV, 7 * u.TeV),
                SpectrumEnergyGroup(2, 6, 8, "normal", 7 * u.TeV, 10 * u.TeV),
            ]
        )

        assert groups == expected

    def test_compute_groups_fixed_outside_range(self, obs):
        ebounds = [20, 30, 40] * u.TeV
        seg = SpectrumEnergyGroupMaker(obs=obs)
        with pytest.raises(ValueError):
            seg.compute_groups_fixed(ebounds=ebounds)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import namedtuple, OrderedDict
import numpy as np
from astropy.table import Table
from astropy.units import Unit, Quantity
from astropy.coordinates import Angle, SkyCoord
from astropy.time import Time
from astropy.utils import lazyproperty
from ..utils.scripts import make_path
from ..utils.time import time_relative_to_ref
from ..utils.testing import Checker
from .gti import GTI

__all__ = ["ObservationTable"]


class ObservationTable(Table):
    """Observation table.

    Data format specification: :ref:`gadf:obs-index`
    """

    @classmethod
    def read(cls, filename, **kwargs):
        """Read an observation table from file.

        Parameters
        ----------
        filename : `~gammapy.extern.pathlib.Path`, str
            Filename
        """
        filename = make_path(filename)
        return super(ObservationTable, cls).read(str(filename), **kwargs)

    @property
    def pointing_radec(self):
        """Pointing positions as ICRS (`~astropy.coordinates.SkyCoord`)"""
        return SkyCoord(self["RA_PNT"], self["DEC_PNT"], unit="deg", frame="icrs")

    @property
    def pointing_galactic(self):
        """Pointing positions as Galactic (`~astropy.coordinates.SkyCoord`)"""
        return SkyCoord(
            self["GLON_PNT"], self["GLAT_PNT"], unit="deg", frame="galactic"
        )

    @lazyproperty
    def _index_dict(self):
        """Dict containing row index for all obs ids"""
        # TODO: Switch to http://docs.astropy.org/en/latest/table/indexing.html once it is more stable
        temp = zip(self["OBS_ID"], np.arange(len(self)))
        return dict(temp)

    def get_obs_idx(self, obs_id):
        """Get row index for given ``obs_id``.

        Raises KeyError if observation is not available.

        Parameters
        ----------
        obs_id : int, list
            observation ids

        Returns
        -------
        idx : list
            indices corresponding to obs_id
        """
        idx = [self._index_dict[key] for key in np.atleast_1d(obs_id)]
        return idx

    def select_obs_id(self, obs_id):
        """Get `~gammapy.data.ObservationTable` containing only ``obs_id``.

        Raises KeyError if observation is not available.

        Parameters
        ----------
        obs_id: int, list
            observation ids
        """
        return self[self.get_obs_idx(obs_id)]

    def summary(self):
        """Info string (str)"""
        obs_name = self.meta.get("OBSERVATORY_NAME", "N/A")

        return "\n".join(
            [
                "Observation table:",
                "Observatory name: {!r}".format(obs_name),
                "Number of observations: {}".format(len(self)),
                # TODO: clean this up. Make those properties?
                # ontime = Quantity(self['ONTIME'].sum(), self['ONTIME'].unit)
                #
                # ss += 'Total observation time: {}\n'.format(ontime)
                # livetime = Quantity(self['LIVETIME'].sum(), self['LIVETIME'].unit)
                # ss += 'Total live time: {}\n'.format(livetime)
                # dtf = 100. * (1 - livetime / ontime)
                # ss += 'Average dead time fraction: {:5.2f}%\n'.format(dtf)
                # time_ref = time_ref_from_dict(self.meta)
                # time_ref_unit = time_ref_from_dict(self.meta).format
                # ss += 'Time reference: {} {}'.format(time_ref, time_ref_unit)
            ]
        )

    def select_linspace_subset(self, num):
        """Select subset of observations.

        This is mostly useful for testing, if you want to make
        the analysis run faster.

        Parameters
        ----------
        num : int
            Number of samples to select.

        Returns
        -------
        table : `ObservationTable`
            Subset observation table (a copy).
        """
        indices = np.linspace(start=0, stop=len(self), num=num, endpoint=False)
        # Round down to nearest integer
        indices = indices.astype("int")
        return self[indices]

    def select_range(self, selection_variable, value_range, inverted=False):
        """Make an observation table, applying some selection.

        Generic function to apply a 1D box selection (min, max) to a
        table on any variable that is in the observation table and can
        be casted into a `~astropy.units.Quantity` object.

        If the range length is 0 (min = max), the selection is applied
        to the exact value indicated by the min value. This is useful
        for selection of exact values, for instance in discrete
        variables like the number of telescopes.

        If the inverted flag is activated, the selection is applied to
        keep all elements outside the selected range.

        Parameters
        ----------
        selection_variable : str
            Name of variable to apply a cut (it should exist on the table).
        value_range : `~astropy.units.Quantity`-like
            Allowed range of values (min, max). The type should be
            consistent with the selection_variable.
        inverted : bool, optional
            Invert selection: keep all entries outside the (min, max) range.

        Returns
        -------
        obs_table : `~gammapy.data.ObservationTable`
            Observation table after selection.
        """
        value_range = Quantity(value_range)

        # read values into a quantity in case units have to be taken into account
        value = Quantity(self[selection_variable])

        mask = (value_range[0] <= value) & (value < value_range[1])

        if np.allclose(value_range[0].value, value_range[1].value):
            mask = value_range[0] == value

        if inverted:
            mask = np.invert(mask)

        return self[mask]

    def select_time_range(self, selection_variable, time_range, inverted=False):
        """Make an observation table, applying a time selection.

        Apply a 1D box selection (min, max) to a
        table on any time variable that is in the observation table.
        It supports both fomats: absolute times in
        `~astropy.time.Time` variables and [MET]_.

        If the inverted flag is activated, the selection is applied to
        keep all elements outside the selected range.

        Parameters
        ----------
        selection_variable : str
            Name of variable to apply a cut (it should exist on the table).
        time_range : `~astropy.time.Time`
            Allowed time range (min, max).
        inverted : bool, optional
            Invert selection: keep all entries outside the (min, max) range.

        Returns
        -------
        obs_table : `~gammapy.data.ObservationTable`
            Observation table after selection.
        """
        if self.meta["TIME_FORMAT"] == "absolute":
            # read times into a Time object
            time = Time(self[selection_variable])
        else:
            # transform time to MET
            time_range = time_relative_to_ref(time_range, self.meta)
            # read values into a quantity in case units have to be taken into account
            time = Quantity(self[selection_variable])

        mask = (time_range[0] <= time) & (time < time_range[1])

        if inverted:
            mask = np.invert(mask)

        return self[mask]

    def select_observations(self, selection=None):
        """Select subset of observations.

        Returns a new observation table representing the subset.

        There are 3 main kinds of selection criteria, according to the
        value of the **type** keyword in the **selection** dictionary:

        - sky regions (boxes or circles)

        - time intervals (min, max)

        - intervals (min, max) on any other parameter present in the
          observation table, that can be casted into an
          `~astropy.units.Quantity` object

        Allowed selection criteria are interpreted using the following
        keywords in the **selection** dictionary under the **type** key.

        - ``sky_box`` and ``sky_circle`` are 2D selection criteria acting
          on sky coordinates

            - ``sky_box`` is a squared region delimited by the **lon** and
              **lat** keywords: both tuples of format (min, max); uses
              `~gammapy.catalog.select_sky_box`

            - ``sky_circle`` is a circular region centered in the coordinate
              marked by the **lon** and **lat** keywords, and radius **radius**;
              uses `~gammapy.catalog.select_sky_circle`

          in each case, the coordinate system can be specified by the **frame**
          keyword (built-in Astropy coordinate frames are supported, e.g.
          ``icrs`` or ``galactic``); an aditional border can be defined using
          the **border** keyword

        - ``time_box`` is a 1D selection criterion acting on the observation
          start time (**TSTART**); the interval is set via the
          **time_range** keyword; uses
          `~gammapy.data.ObservationTable.select_time_range`

        - ``par_box`` is a 1D selection criterion acting on any
          parameter defined in the observation table that can be casted
          into an `~astropy.units.Quantity` object; the parameter name
          and interval can be specified using the keywords **variable** and
          **value_range** respectively; min = max selects exact
          values of the parameter; uses
          `~gammapy.data.ObservationTable.select_range`

        In all cases, the selection can be inverted by activating the
        **inverted** flag, in which case, the selection is applied to keep all
        elements outside the selected range.

        A few examples of selection criteria are given below.

        Parameters
        ----------
        selection : dict
            Dictionary with a few keywords for applying selection cuts.

        Returns
        -------
        obs_table : `~gammapy.data.ObservationTable`
            Observation table after selection.

        Examples
        --------
        >>> selection = dict(type='sky_box', frame='icrs',
        ...                  lon=Angle([150, 300], 'deg'),
        ...                  lat=Angle([-50, 0], 'deg'),
        ...                  border=Angle(2, 'deg'))
        >>> selected_obs_table = obs_table.select_observations(selection)

        >>> selection = dict(type='sky_circle', frame='galactic',
        ...                  lon=Angle(0, 'deg'),
        ...                  lat=Angle(0, 'deg'),
        ...                  radius=Angle(5, 'deg'),
        ...                  border=Angle(2, 'deg'))
        >>> selected_obs_table = obs_table.select_observations(selection)

        >>> selection = dict(type='time_box',
        ...                  time_range=Time(['2012-01-01T01:00:00', '2012-01-01T02:00:00']))
        >>> selected_obs_table = obs_table.select_observations(selection)

        >>> selection = dict(type='par_box', variable='ALT',
        ...                  value_range=Angle([60., 70.], 'deg'))
        >>> selected_obs_table = obs_table.select_observations(selection)

        >>> selection = dict(type='par_box', variable='OBS_ID',
        ...                  value_range=[2, 5])
        >>> selected_obs_table = obs_table.select_observations(selection)

        >>> selection = dict(type='par_box', variable='N_TELS',
        ...                  value_range=[4, 4])
        >>> selected_obs_table = obs_table.select_observations(selection)
        """
        from ..catalog import select_sky_box, select_sky_circle

        if "inverted" not in selection.keys():
            selection["inverted"] = False

        if selection["type"] == "sky_circle":
            lon = selection["lon"]
            lat = selection["lat"]
            radius = selection["radius"] + selection["border"]
            return select_sky_circle(
                self,
                lon_cen=lon,
                lat_cen=lat,
                radius=radius,
                frame=selection["frame"],
                inverted=selection["inverted"],
            )

        elif selection["type"] == "sky_box":
            lon = selection["lon"]
            lat = selection["lat"]
            border = selection["border"]
            lon = Angle([lon[0] - border, lon[1] + border])
            lat = Angle([lat[0] - border, lat[1] + border])
            return select_sky_box(
                self,
                lon_lim=lon,
                lat_lim=lat,
                frame=selection["frame"],
                inverted=selection["inverted"],
            )

        elif selection["type"] == "time_box":
            return self.select_time_range(
                "TSTART", selection["time_range"], selection["inverted"]
            )

        elif selection["type"] == "par_box":
            return self.select_range(
                selection["variable"], selection["value_range"], selection["inverted"]
            )

        else:
            raise ValueError("Invalid selection type: {}".format(selection["type"]))

    def create_gti(self, obs_id):
        """Returns a GTI table containing TSTART and TSTOP from the table.

        TODO: This method can be removed once GTI tables are explicitly required in Gammapy.

        Parameters
        ----------
        obs_id : int
            ID of the observation for which the GTI table will be created

        Returns
        -------
        gti : `~gammapy.data.GTI`
            GTI table containing one row (TSTART and TSTOP of the observation with `obs_id`)
        """
        meta = OrderedDict(
            EXTNAME="GTI",
            HDUCLASS="GADF",
            HDUDOC="https://github.com/open-gamma-ray-astro/gamma-astro-data-formats",
            HDUVERS="0.2",
            HDUCLAS1="GTI",
            MJDREFI=self.meta["MJDREFI"],
            MJDREFF=self.meta["MJDREFF"],
            TIMEUNIT=self.meta.get("TIMEUNIT", "s"),
            TIMESYS=self.meta["TIMESYS"],
            TIMEREF=self.meta.get("TIMEREF", "LOCAL"),
        )

        obs = self.select_obs_id(obs_id)

        gti = Table(meta=meta)
        gti["START"] = obs["TSTART"].quantity.to("s")
        gti["STOP"] = obs["TSTOP"].quantity.to("s")

        return GTI(gti)


class ObservationTableChecker(Checker):
    """Event list checker.

    Data format specification: ref:`gadf:iact-events`

    Parameters
    ----------
    event_list : `~gammapy.data.EventList`
        Event list
    """

    CHECKS = {
        "meta": "check_meta",
        "columns": "check_columns",
        # "times": "check_times",
        # "coordinates_galactic": "check_coordinates_galactic",
        # "coordinates_altaz": "check_coordinates_altaz",
    }

    # accuracy = {"angle": Angle("1 arcsec"), "time": Quantity(1, "microsecond")}

    # https://gamma-astro-data-formats.readthedocs.io/en/latest/events/events.html#mandatory-header-keywords
    meta_required = [
        "HDUCLASS",
        "HDUDOC",
        "HDUVERS",
        "HDUCLAS1",
        "HDUCLAS2",
        # https://gamma-astro-data-formats.readthedocs.io/en/latest/general/time.html#time-formats
        "MJDREFI",
        "MJDREFF",
        "TIMEUNIT",
        "TIMESYS",
        "TIMEREF",
        # https://gamma-astro-data-formats.readthedocs.io/en/latest/general/coordinates.html#coords-location
        "GEOLON",
        "GEOLAT",
        "ALTITUDE",
    ]

    _col = namedtuple("col", ["name", "unit"])
    columns_required = [
        _col(name="OBS_ID", unit=""),
        _col(name="RA_PNT", unit="deg"),
        _col(name="DEC_PNT", unit="deg"),
        _col(name="TSTART", unit="s"),
        _col(name="TSTOP", unit="s"),
    ]

    def __init__(self, obs_table):
        self.obs_table = obs_table

    def _record(self, level="info", msg=None):
        return {"level": level, "hdu": "obs-index", "msg": msg}

    def check_meta(self):
        m = self.obs_table.meta

        meta_missing = sorted(set(self.meta_required) - set(m))
        if meta_missing:
            yield self._record(
                level="error", msg="Missing meta keys: {!r}".format(meta_missing)
            )

        if m.get("HDUCLAS1", "") != "INDEX":
            yield self._record(level="error", msg="HDUCLAS1 must be INDEX")
        if m.get("HDUCLAS2", "") != "OBS":
            yield self._record(level="error", msg="HDUCLAS2 must be OBS")

    def check_columns(self):
        t = self.obs_table

        if len(t) == 0:
            yield self._record(level="error", msg="Observation table has zero rows")

        for name, unit in self.columns_required:
            if name not in t.colnames:
                yield self._record(
                    level="error", msg="Missing table column: {!r}".format(name)
                )
            else:
                if Unit(unit) != (t[name].unit or ""):
                    yield self._record(
                        level="error", msg="Invalid unit for column: {!r}".format(name)
                    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import copy
import operator
import numpy as np
from astropy.utils.decorators import lazyproperty
import astropy.units as u
from ..utils.fitting import Parameters, Parameter, Model
from ..utils.scripts import make_path
from ..maps import Map

__all__ = [
    "SkyModelBase",
    "SkyModels",
    "SkyModel",
    "CompoundSkyModel",
    "SkyDiffuseCube",
    "BackgroundModel",
]


class SkyModelBase(Model):
    """Sky model base class"""

    def __add__(self, skymodel):
        return CompoundSkyModel(self, skymodel, operator.add)

    def __radd__(self, model):
        return self.__add__(model)

    def __call__(self, lon, lat, energy):
        return self.evaluate(lon, lat, energy)


class SkyModels(Model):
    """Collection of `~gammapy.cube.models.SkyModel`

    Parameters
    ----------
    skymodels : list of `~gammapy.cube.models.SkyModel`
        Sky models

    Examples
    --------

    Read from an XML file::

        from gammapy.cube import SkyModels
        filename = '$GAMMAPY_DATA/tests/models/fermi_model.xml'
        sourcelib = SkyModels.read(filename)
    """

    def __init__(self, skymodels):
        self.skymodels = skymodels

        pars = []
        for skymodel in skymodels:
            for p in skymodel.parameters:
                pars.append(p)
        self._parameters = Parameters(pars)

    @property
    def parameters(self):
        """Concatenated parameters.

        Currently no way to distinguish spectral and spatial.
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        idx = 0
        for skymodel in self.skymodels:
            n_par = len(skymodel.parameters.parameters)
            skymodel.parameters.parameters = parameters.parameters[idx : idx + n_par]
            idx += n_par

    @classmethod
    def from_xml(cls, xml):
        """Read from XML string."""
        from ..utils.serialization import xml_to_sky_models

        return xml_to_sky_models(xml)

    @classmethod
    def read(cls, filename):
        """Read from XML file.

        The XML definition of some models is uncompatible with the models
        currently implemented in gammapy. Therefore the following modifications
        happen to the XML model definition

        * PowerLaw: The spectral index is negative in XML but positive in
          gammapy. Parameter limits are ignored

        * ExponentialCutoffPowerLaw: The cutoff energy is transferred to
          lambda = 1 / cutof energy on read
        """
        path = make_path(filename)
        xml = path.read_text()
        return cls.from_xml(xml)

    def to_xml(self, filename):
        """Write to XML file."""
        from ..utils.serialization import sky_models_to_xml

        xml = sky_models_to_xml(self)
        filename = make_path(filename)
        with filename.open("w") as output:
            output.write(xml)

    def to_compound_model(self):
        """Return `~gammapy.cube.models.CompoundSkyModel`"""
        return np.sum([m for m in self.skymodels])

    def evaluate(self, lon, lat, energy):
        out = self.skymodels[0].evaluate(lon, lat, energy)
        for skymodel in self.skymodels[1:]:
            out += skymodel.evaluate(lon, lat, energy)
        return out


class SkyModel(SkyModelBase):
    """Sky model component.

    This model represents a factorised sky model.
    It has a `~gammapy.utils.modeling.Parameters`
    combining the spatial and spectral parameters.

    TODO: add possibility to have a temporal model component also.

    Parameters
    ----------
    spatial_model : `~gammapy.image.models.SkySpatialModel`
        Spatial model (must be normalised to integrate to 1)
    spectral_model : `~gammapy.spectrum.models.SpectralModel`
        Spectral model
    name : str
        Model identifier
    """

    def __init__(self, spatial_model, spectral_model, name="SkyModel"):
        self.name = name
        self._spatial_model = spatial_model
        self._spectral_model = spectral_model
        self._parameters = Parameters(
            spatial_model.parameters.parameters + spectral_model.parameters.parameters
        )

    @property
    def spatial_model(self):
        """`~gammapy.image.models.SkySpatialModel`"""
        # propagate sub-covariance
        if self.parameters.covariance is not None:
            idx = len(self._spatial_model.parameters.parameters)
            self._spatial_model.parameters.covariance = self.parameters.covariance[
                :idx, :idx
            ]
        return self._spatial_model

    @property
    def spectral_model(self):
        """`~gammapy.spectrum.models.SpectralModel`"""
        # propagate sub-covariance
        if self.parameters.covariance is not None:
            idx = len(self._spatial_model.parameters.parameters)
            self._spectral_model.parameters.covariance = self.parameters.covariance[
                idx:, idx:
            ]
        return self._spectral_model

    @property
    def parameters(self):
        """Parameters (`~gammapy.utils.modeling.Parameters`)"""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = parameters
        idx = len(self.spatial_model.parameters.parameters)
        self._spatial_model.parameters.parameters = parameters.parameters[:idx]
        self._spectral_model.parameters.parameters = parameters.parameters[idx:]

    def __repr__(self):
        fmt = "{}(spatial_model={!r}, spectral_model={!r})"
        return fmt.format(
            self.__class__.__name__, self.spatial_model, self.spectral_model
        )

    def evaluate(self, lon, lat, energy):
        """Evaluate the model at given points.

        The model evaluation follows numpy broadcasting rules.

        Return differential surface brightness cube.
        At the moment in units: ``cm-2 s-1 TeV-1 deg-2``

        Parameters
        ----------
        lon, lat : `~astropy.units.Quantity`
            Spatial coordinates
        energy : `~astropy.units.Quantity`
            Energy coordinate

        Returns
        -------
        value : `~astropy.units.Quantity`
            Model value at the given point.
        """
        val_spatial = self.spatial_model(lon, lat)  # pylint:disable=not-callable
        val_spectral = self.spectral_model(energy)  # pylint:disable=not-callable
        val = val_spatial * val_spectral
        # TODO: shall remove hard coded return units? If really needed users can
        # always do this themselves. For fitting this also adds a performance penalty...
        return val.to("cm-2 s-1 TeV-1 deg-2")


class CompoundSkyModel(SkyModelBase):
    """Represents the algebraic combination of two
    `~gammapy.cube.models.SkyModel`

    Parameters
    ----------
    model1, model2 : `SkyModel`
        Two sky models
    operator : callable
        Binary operator to combine the models
    """

    def __init__(self, model1, model2, operator):
        self.model1 = model1
        self.model2 = model2
        self.operator = operator
        self._parameters = Parameters(
            self.model1.parameters.parameters + self.model2.parameters.parameters
        )

    @property
    def parameters(self):
        """Parameters (`~gammapy.utils.modeling.Parameters`)"""
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = parameters
        idx = len(self.model1.parameters.parameters)
        self.model1.parameters.parameters = parameters.parameters[:idx]
        self.model2.parameters.parameters = parameters.parameters[idx:]

    def __str__(self):
        ss = self.__class__.__name__
        ss += "\n    Component 1 : {}".format(self.model1)
        ss += "\n    Component 2 : {}".format(self.model2)
        ss += "\n    Operator : {}".format(self.operator)
        return ss

    def evaluate(self, lon, lat, energy):
        """Evaluate the compound model at given points.

        Return differential surface brightness cube.
        At the moment in units: ``cm-2 s-1 TeV-1 deg-2``

        Parameters
        ----------
        lon, lat : `~astropy.units.Quantity`
            Spatial coordinates
        energy : `~astropy.units.Quantity`
            Energy coordinate

        Returns
        -------
        value : `~astropy.units.Quantity`
            Model value at the given point.
        """
        val1 = self.model1.evaluate(lon, lat, energy)
        val2 = self.model2.evaluate(lon, lat, energy)
        return self.operator(val1, val2)


class SkyDiffuseCube(SkyModelBase):
    """Cube sky map template model (3D).

    This is for a 3D map with an energy axis. Use `~gammapy.image.models.SkyDiffuseMap`
    for 2D maps.

    Parameters
    ----------
    map : `~gammapy.maps.Map`
        Map template
    norm : float
        Norm parameter (multiplied with map values)
    meta : dict, optional
        Meta information, meta['filename'] will be used for serialization
    interp_kwargs : dict
        Interpolation keyword arguments passed to `Map.interp_by_coord()`.
        Default arguments are {'interp': 'linear', 'fill_value': 0}.

    """

    def __init__(self, map, norm=1, meta=None, interp_kwargs=None):
        axis = map.geom.get_axis_by_name("energy")

        if axis.node_type != "center":
            raise ValueError('Need a map with energy axis node_type="center"')

        self.map = map
        self.parameters = Parameters([Parameter("norm", norm)])
        self.meta = {} if meta is None else meta

        interp_kwargs = {} if interp_kwargs is None else interp_kwargs
        interp_kwargs.setdefault("interp", "linear")
        interp_kwargs.setdefault("fill_value", 0)
        self._interp_kwargs = interp_kwargs

    @classmethod
    def read(cls, filename, **kwargs):
        """Read map from FITS file.

        The default unit used if none is found in the file is ``cm-2 s-1 MeV-1 sr-1``.

        Parameters
        ----------
        filename : str
            FITS image filename.
        """
        m = Map.read(filename, **kwargs)
        if m.unit == "":
            m.unit = "cm-2 s-1 MeV-1 sr-1"
        return cls(m)

    def evaluate(self, lon, lat, energy):
        """Evaluate model."""
        coord = {
            "lon": lon.to_value("deg"),
            "lat": lat.to_value("deg"),
            "energy": energy,
        }
        val = self.map.interp_by_coord(coord, **self._interp_kwargs)
        norm = self.parameters["norm"].value
        return u.Quantity(norm * val, self.map.unit, copy=False)

    def copy(self):
        """A shallow copy"""
        return copy.copy(self)


class BackgroundModel(Model):
    """Background model

    Parameters
    ----------
    background : `~gammapy.maps.Map`
        Background model map
    norm : float
        Background normalisation
    tilt : float
        Additional tilt in the spectrum
    reference : `~astropy.units.Quantity`
        Reference energy of the tilt.
    """

    def __init__(self, background, norm=1, tilt=0, reference="1 TeV"):

        axis = background.geom.get_axis_by_name("energy")
        if axis.node_type != "edges":
            raise ValueError('Need an integrated map, energy axis node_type="edges"')

        self.map = background
        self.parameters = Parameters(
            [
                Parameter("norm", norm, unit=""),
                Parameter("tilt", tilt, unit="", frozen=True),
                Parameter("reference", reference, frozen=True),
            ]
        )

    @lazyproperty
    def energy_center(self):
        """True energy axis bin centers (`~astropy.units.Quantity`)"""
        energy_axis = self.map.geom.get_axis_by_name("energy")
        energy = energy_axis.center * energy_axis.unit
        return energy[:, np.newaxis, np.newaxis]

    def evaluate(self):
        """Evaluate background model"""
        norm = self.parameters["norm"].value
        tilt = self.parameters["tilt"].value
        reference = self.parameters["reference"].quantity
        tilt_factor = np.power((self.energy_center / reference).to(""), -tilt)
        return u.Quantity(norm * self.map.data * tilt_factor, self.map.unit, copy=False)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
from numpy.testing import assert_allclose
from astropy.table import Table
from ...utils.scripts import make_path
from ...utils.testing import requires_data
from ..models import PhaseCurveTableModel, LightCurveTableModel


@pytest.fixture(scope="session")
def phase_curve():
    filename = make_path("$GAMMAPY_DATA/tests/phasecurve_LSI_DC.fits")
    table = Table.read(str(filename))
    return PhaseCurveTableModel(
        table, time_0=43366.275, phase_0=0.0, f0=4.367575e-7, f1=0.0, f2=0.0
    )


@requires_data("gammapy-data")
def test_phasecurve_phase(phase_curve):
    time = 46300.0
    phase = phase_curve.phase(time)
    assert_allclose(phase, 0.7066006737999402)


@requires_data("gammapy-data")
def test_phasecurve_evaluate(phase_curve):
    time = 46300.0
    value = phase_curve.evaluate_norm_at_time(time)
    assert_allclose(value, 0.49059393580053845)

    phase = phase_curve.phase(time)
    value = phase_curve.evaluate_norm_at_phase(phase)
    assert_allclose(value, 0.49059393580053845)


# TODO: add light-curve test case from scratch
# only use the FITS one for I/O (or not at all)
@pytest.fixture(scope="session")
def light_curve():
    path = "$GAMMAPY_DATA/tests/models/light_curve/lightcrv_PKSB1222+216.fits"
    return LightCurveTableModel.read(path)


@requires_data("gammapy-data")
def test_light_curve_str(light_curve):
    ss = str(light_curve)
    assert "LightCurveTableModel" in ss


@requires_data("gammapy-data")
def test_light_curve_evaluate_norm_at_time(light_curve):
    val = light_curve.evaluate_norm_at_time(46300)
    assert_allclose(val, 0.021192223042749835)


@requires_data("gammapy-data")
def test_light_curve_mean_norm_in_time_interval(light_curve):
    val = light_curve.mean_norm_in_time_interval(46300, 46301)
    assert_allclose(val, 0.021192284384617066)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from ...utils.testing import requires_data
from ...maps import MapAxis, WcsGeom, Map
from ...irf.energy_dispersion import EnergyDispersion
from ...cube.psf_kernel import PSFKernel
from ...cube.models import SkyDiffuseCube, BackgroundModel
from ...image.models import SkyGaussian
from ...spectrum.models import PowerLaw
from ..fit import MapEvaluator
from ..models import SkyModel, SkyModels, CompoundSkyModel


@pytest.fixture(scope="session")
def sky_model():
    spatial_model = SkyGaussian(lon_0="3 deg", lat_0="4 deg", sigma="3 deg")
    spectral_model = PowerLaw(
        index=2, amplitude="1e-11 cm-2 s-1 TeV-1", reference="1 TeV"
    )
    return SkyModel(spatial_model, spectral_model)


@pytest.fixture(scope="session")
def diffuse_model():
    axis = MapAxis.from_nodes([0.1, 100], name="energy", unit="TeV", interp="log")
    m = Map.create(npix=(4, 3), binsz=2, axes=[axis], unit="cm-2 s-1 MeV-1 sr-1")
    m.data += 42
    return SkyDiffuseCube(m)


@pytest.fixture(scope="session")
def geom():
    axis = MapAxis.from_edges(np.logspace(-1, 1, 3), unit=u.TeV, name="energy")
    return WcsGeom.create(skydir=(0, 0), npix=(5, 4), coordsys="GAL", axes=[axis])


@pytest.fixture(scope="session")
def geom_true():
    axis = MapAxis.from_edges(np.logspace(-1, 1, 4), unit=u.TeV, name="energy")
    return WcsGeom.create(skydir=(0, 0), npix=(5, 4), coordsys="GAL", axes=[axis])


@pytest.fixture(scope="session")
def exposure(geom_true):
    m = Map.from_geom(geom_true)
    m.quantity = np.ones(geom_true.data_shape) * u.Quantity("100 m2 s")
    m.data[1] *= 10
    return m


@pytest.fixture(scope="session")
def background(geom):
    m = Map.from_geom(geom)
    m.quantity = np.ones(geom.data_shape) * 1e-7
    return m


@pytest.fixture(scope="session")
def edisp(geom, geom_true):
    e_reco = geom.get_axis_by_name("energy").edges
    e_true = geom_true.get_axis_by_name("energy").edges
    return EnergyDispersion.from_diagonal_response(e_true=e_true, e_reco=e_reco)


@pytest.fixture(scope="session")
def psf(geom_true):
    sigma = 0.5 * u.deg
    return PSFKernel.from_gauss(geom_true, sigma)


@pytest.fixture(scope="session")
def evaluator(sky_model, exposure, background, psf, edisp):
    return MapEvaluator(sky_model, exposure, background, psf=psf, edisp=edisp)


@pytest.fixture(scope="session")
def diffuse_evaluator(diffuse_model, exposure, background, psf, edisp):
    return MapEvaluator(diffuse_model, exposure, background, psf=psf, edisp=edisp)


@pytest.fixture(scope="session")
def sky_models(sky_model):
    return SkyModels([sky_model, sky_model])


@pytest.fixture(scope="session")
def compound_model(sky_model):
    return sky_model + sky_model


def test_background_model(background):
    bkg1 = BackgroundModel(background, norm=2.0).evaluate()
    assert_allclose(bkg1[0][0][0], background.data[0][0][0] * 2.0, rtol=1e-3)
    assert_allclose(bkg1.sum(), background.data.sum() * 2.0, rtol=1e-3)

    bkg2 = BackgroundModel(
        background, norm=2.0, tilt=0.2, reference="1000 GeV"
    ).evaluate()
    assert_allclose(bkg2[0][0][0], 2.254e-07, rtol=1e-3)
    assert_allclose(bkg2.sum(), 7.352e-06, rtol=1e-3)


class TestSkyModels:
    @staticmethod
    def test_to_compound_model(sky_models):
        model = sky_models.to_compound_model()
        assert isinstance(model, CompoundSkyModel)
        pars = model.parameters.parameters
        assert len(pars) == 12
        assert pars[0].name == "lon_0"
        assert pars[-1].name == "reference"

    @staticmethod
    def test_parameters(sky_models):
        parnames = ["lon_0", "lat_0", "sigma", "index", "amplitude", "reference"] * 2
        assert sky_models.parameters.names == parnames

        # Check that model parameters are references to the parts
        p1 = sky_models.parameters["lon_0"]
        p2 = sky_models.skymodels[0].parameters["lon_0"]
        assert p1 is p2

    @staticmethod
    def test_evaluate(sky_models):
        lon = 3 * u.deg * np.ones(shape=(3, 4))
        lat = 4 * u.deg * np.ones(shape=(3, 4))
        energy = [1, 1, 1, 1, 1] * u.TeV

        q = sky_models.evaluate(lon, lat, energy[:, np.newaxis, np.newaxis])

        assert q.unit == "cm-2 s-1 TeV-1 deg-2"
        assert q.shape == (5, 3, 4)
        assert_allclose(q.value, 3.536776513153229e-13)


class TestSkyModel:
    @staticmethod
    def test_repr(sky_model):
        assert "SkyModel" in repr(sky_model)

    @staticmethod
    def test_str(sky_model):
        assert "SkyModel" in str(sky_model)

    @staticmethod
    def test_parameters(sky_model):
        # Check that model parameters are references to the spatial and spectral parts
        p1 = sky_model.parameters["lon_0"]
        p2 = sky_model.spatial_model.parameters["lon_0"]
        assert p1 is p2

        p1 = sky_model.parameters["amplitude"]
        p2 = sky_model.spectral_model.parameters["amplitude"]
        assert p1 is p2

    @staticmethod
    def test_evaluate_scalar(sky_model):
        lon = 3 * u.deg
        lat = 4 * u.deg
        energy = 1 * u.TeV

        q = sky_model.evaluate(lon, lat, energy)

        assert q.unit == "cm-2 s-1 TeV-1 deg-2"
        assert np.isscalar(q.value)
        assert_allclose(q.value, 1.76838826e-13)

    @staticmethod
    def test_evaluate_array(sky_model):
        lon = 3 * u.deg * np.ones(shape=(3, 4))
        lat = 4 * u.deg * np.ones(shape=(3, 4))
        energy = [1, 1, 1, 1, 1] * u.TeV

        q = sky_model.evaluate(lon, lat, energy[:, np.newaxis, np.newaxis])

        assert q.shape == (5, 3, 4)
        assert_allclose(q.value, 1.76838826e-13)


class TestCompoundSkyModel:
    @staticmethod
    def test_parameters(compound_model):
        parnames = ["lon_0", "lat_0", "sigma", "index", "amplitude", "reference"] * 2
        assert compound_model.parameters.names == parnames

        # Check that model parameters are references to the parts
        assert (
            compound_model.parameters["lon_0"]
            is compound_model.model1.parameters["lon_0"]
        )

        # Check that parameter assignment works
        assert compound_model.parameters.parameters[-1].value == 1
        compound_model.parameters = compound_model.parameters.copy()
        assert compound_model.parameters.parameters[-1].value == 1

    @staticmethod
    def test_evaluate(compound_model):
        lon = 3 * u.deg * np.ones(shape=(3, 4))
        lat = 4 * u.deg * np.ones(shape=(3, 4))
        energy = [1, 1, 1, 1, 1] * u.TeV

        q = compound_model.evaluate(lon, lat, energy[:, np.newaxis, np.newaxis])

        assert q.unit == "cm-2 s-1 TeV-1 deg-2"
        assert q.shape == (5, 3, 4)
        assert_allclose(q.value, 3.536776513153229e-13)


class TestSkyDiffuseCube:
    @staticmethod
    def test_evaluate_scalar(diffuse_model):
        # Check pixel inside map
        val = diffuse_model.evaluate(0 * u.deg, 0 * u.deg, 10 * u.TeV)
        assert val.unit == "cm-2 s-1 MeV-1 sr-1"
        assert val.shape == (1,)
        assert_allclose(val.value, 42)

        # Check pixel outside map (spatially)
        val = diffuse_model.evaluate(100 * u.deg, 0 * u.deg, 10 * u.TeV)
        assert_allclose(val.value, 0)

        # Check pixel outside energy range
        val = diffuse_model.evaluate(0 * u.deg, 0 * u.deg, 200 * u.TeV)
        assert_allclose(val.value, 0)

    @staticmethod
    def test_evaluate_array(diffuse_model):
        lon = 1 * u.deg * np.ones(shape=(3, 4))
        lat = 2 * u.deg * np.ones(shape=(3, 4))
        energy = [1, 1, 1, 1, 1] * u.TeV

        q = diffuse_model.evaluate(lon, lat, energy[:, np.newaxis, np.newaxis])

        assert q.shape == (5, 3, 4)
        assert_allclose(q.value.mean(), 42)

    @staticmethod
    @requires_data("gammapy-data")
    def test_read():
        model = SkyDiffuseCube.read(
            "$GAMMAPY_DATA/tests/unbundled/fermi/gll_iem_v02_cutout.fits"
        )
        assert model.map.unit == "cm-2 s-1 MeV-1 sr-1"

        # Check pixel inside map
        val = model.evaluate(0 * u.deg, 0 * u.deg, 100 * u.GeV)
        assert val.unit == "cm-2 s-1 MeV-1 sr-1"
        assert val.shape == (1,)
        assert_allclose(val.value, 1.396424e-12, rtol=1e-5)


class TestSkyDiffuseCubeMapEvaluator:
    @staticmethod
    def test_compute_dnde(diffuse_evaluator):
        out = diffuse_evaluator.compute_dnde()
        assert out.shape == (3, 4, 5)
        out = out.to("cm-2 s-1 MeV-1 sr-1")
        assert_allclose(out.value.sum(), 2520.0, rtol=1e-5)
        assert_allclose(out.value[0, 0, 0], 42, rtol=1e-5)

    @staticmethod
    def test_compute_flux(diffuse_evaluator):
        out = diffuse_evaluator.compute_flux()
        assert out.shape == (3, 4, 5)
        out = out.to("cm-2 s-1")
        assert_allclose(out.value.sum(), 633263.444803, rtol=1e-5)
        assert_allclose(out.value[0, 0, 0], 1164.578565, rtol=1e-5)

    @staticmethod
    def test_apply_psf(diffuse_evaluator):
        flux = diffuse_evaluator.compute_flux()
        npred = diffuse_evaluator.apply_exposure(flux)
        out = diffuse_evaluator.apply_psf(npred)
        assert out.data.shape == (3, 4, 5)
        assert_allclose(out.data.sum(), 1.106404e12, rtol=1e-5)
        assert_allclose(out.data[0, 0, 0], 5.586252e08, rtol=1e-5)

    @staticmethod
    def test_apply_edisp(diffuse_evaluator):
        flux = diffuse_evaluator.compute_flux()
        npred = diffuse_evaluator.apply_exposure(flux)
        out = diffuse_evaluator.apply_edisp(npred)
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 1.606345e12, rtol=1e-5)
        assert_allclose(out.data[0, 0, 0], 1.164579e09, rtol=1e-5)

    @staticmethod
    def test_compute_npred(diffuse_evaluator):
        out = diffuse_evaluator.compute_npred()
        assert out.shape == (2, 4, 5)
        assert_allclose(out.sum(), 1.106403e12, rtol=1e-5)
        assert_allclose(out[0, 0, 0], 5.586252e08, rtol=1e-5)


class TestSkyModelMapEvaluator:
    @staticmethod
    def test_energy_center(evaluator):
        val = evaluator.energy_center
        assert val.shape == (3, 1, 1)
        assert val.unit == "TeV"

    @staticmethod
    def test_energy_edges(evaluator):
        val = evaluator.energy_edges
        assert val.shape == (4, 1, 1)
        assert val.unit == "TeV"

    @staticmethod
    def test_energy_bin_width(evaluator):
        val = evaluator.energy_bin_width
        assert val.shape == (3, 1, 1)
        assert val.unit == "TeV"

    @staticmethod
    def test_lon_lat(evaluator):
        val = evaluator.lon
        assert val.shape == (4, 5)
        assert val.unit == "deg"

        val = evaluator.lat
        assert val.shape == (4, 5)
        assert val.unit == "deg"

    @staticmethod
    def test_solid_angle(evaluator):
        val = evaluator.solid_angle
        assert val.shape == (3, 4, 5)
        assert val.unit == "sr"

    @staticmethod
    def test_bin_volume(evaluator):
        val = evaluator.bin_volume
        assert val.shape == (3, 4, 5)
        assert val.unit == "TeV sr"

    @staticmethod
    def test_compute_dnde(evaluator):
        out = evaluator.compute_dnde()
        assert out.shape == (3, 4, 5)
        assert out.unit == "cm-2 s-1 TeV-1 deg-2"
        assert_allclose(out.value.sum(), 1.177907e-11, rtol=1e-5)
        assert_allclose(out.value[0, 0, 0], 5.082551e-13, rtol=1e-5)

    @staticmethod
    def test_compute_flux(evaluator):
        out = evaluator.compute_flux().to_value("cm-2 s-1")
        assert out.shape == (3, 4, 5)
        assert_allclose(out.sum(), 1.290431e-12, rtol=1e-5)
        assert_allclose(out[0, 0, 0], 4.626436e-14, rtol=1e-5)

    @staticmethod
    def test_apply_psf(evaluator):
        flux = evaluator.compute_flux()
        npred = evaluator.apply_exposure(flux)
        out = evaluator.apply_psf(npred)
        assert out.data.shape == (3, 4, 5)
        assert_allclose(out.data.sum(), 2.25133e-06, rtol=1e-5)
        assert_allclose(out.data[0, 0, 0], 2.405235e-08, rtol=1e-5)

    @staticmethod
    def test_apply_edisp(evaluator):
        flux = evaluator.compute_flux()
        npred = evaluator.apply_exposure(flux)
        out = evaluator.apply_edisp(npred)
        assert out.data.shape == (2, 4, 5)
        assert_allclose(out.data.sum(), 3.273326e-06, rtol=1e-5)
        assert_allclose(out.data[0, 0, 0], 4.626436e-08, rtol=1e-5)

    @staticmethod
    def test_compute_npred(evaluator):
        out = evaluator.compute_npred()
        assert out.shape == (2, 4, 5)
        assert_allclose(out.sum(), 6.25133e-06, rtol=1e-5)
        assert_allclose(out[0, 0, 0], 1.240524e-07, rtol=1e-5)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import astropy.units as u
from numpy.testing import assert_allclose
from ...utils.testing import assert_quantity_allclose
from ...utils.testing import requires_dependency, requires_data, mpl_plot_check
from ...utils.scripts import make_path
from ...irf import EffectiveAreaTable, EnergyDispersion
from .. import (
    PHACountsSpectrum,
    SpectrumObservation,
    SpectrumObservationList,
    SpectrumObservationStacker,
    models,
    SpectrumSimulation,
)


@requires_dependency("sherpa")
@requires_data("gammapy-data")
def test_spectrum_observation_1():
    """Obs read from file"""
    filename = "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
    obs = SpectrumObservation.read(filename)
    pars = dict(
        total_on=189,
        livetime=1581.73681640625 * u.second,
        npred=109.014552,
        excess=169.916667,
        excess_safe_range=116.33,
    )
    tester = SpectrumObservationTester(obs, pars)
    tester.test_all()


@requires_dependency("sherpa")
@requires_data("gammapy-data")
def test_spectrum_observation_2():
    """Simulated obs without background"""
    energy = np.logspace(-2, 2, 100) * u.TeV
    aeff = EffectiveAreaTable.from_parametrization(energy=energy)
    edisp = EnergyDispersion.from_gauss(e_true=energy, e_reco=energy, sigma=0.2, bias=0)
    livetime = 1 * u.h
    source_model = models.PowerLaw(
        index=2.3, amplitude=2.3e-11 * u.Unit("cm-2 s-1 TeV-1"), reference=1.4 * u.TeV
    )
    sim = SpectrumSimulation(
        aeff=aeff, edisp=edisp, source_model=source_model, livetime=livetime
    )
    sim.simulate_obs(seed=2309, obs_id=2309)
    obs = sim.obs

    pars = dict(
        total_on=824,
        livetime=livetime,
        npred=292.00223031875987,
        excess=824,
        excess_safe_range=824,
    )
    tester = SpectrumObservationTester(obs, pars)
    tester.test_all()


@requires_dependency("sherpa")
@requires_data("gammapy-data")
def test_spectrum_observation_3():
    """obs without edisp"""
    energy = np.logspace(-1, 1, 20) * u.TeV
    livetime = 2 * u.h
    on_vector = PHACountsSpectrum(
        energy_lo=energy[:-1], energy_hi=energy[1:], data=np.arange(19), backscal=1
    )
    on_vector.livetime = livetime
    on_vector.obs_id = 2
    aeff = EffectiveAreaTable(
        energy_lo=energy[:-1], energy_hi=energy[1:], data=np.ones(19) * 1e5 * u.m ** 2
    )
    obs = SpectrumObservation(on_vector=on_vector, aeff=aeff)
    pars = dict(
        total_on=171, livetime=livetime, npred=1425.6, excess=171, excess_safe_range=171
    )
    tester = SpectrumObservationTester(obs, pars)
    tester.test_all()


@requires_dependency("sherpa")
@requires_data("gammapy-data")
def make_observation_list():
    """obs with dummy IRF"""
    nbin = 3
    energy = np.logspace(-1, 1, nbin + 1) * u.TeV
    livetime = 2 * u.h
    data_on = np.arange(nbin)
    dataoff_1 = np.ones(3)
    dataoff_2 = np.ones(3) * 3
    dataoff_1[1] = 0
    dataoff_2[1] = 0
    on_vector = PHACountsSpectrum(
        energy_lo=energy[:-1], energy_hi=energy[1:], data=data_on, backscal=1
    )
    off_vector1 = PHACountsSpectrum(
        energy_lo=energy[:-1], energy_hi=energy[1:], data=dataoff_1, backscal=2
    )
    off_vector2 = PHACountsSpectrum(
        energy_lo=energy[:-1], energy_hi=energy[1:], data=dataoff_2, backscal=4
    )
    aeff = EffectiveAreaTable(
        energy_lo=energy[:-1], energy_hi=energy[1:], data=np.ones(nbin) * 1e5 * u.m ** 2
    )
    edisp = EnergyDispersion.from_gauss(e_true=energy, e_reco=energy, sigma=0.2, bias=0)
    on_vector.livetime = livetime
    on_vector.obs_id = 2
    obs1 = SpectrumObservation(
        on_vector=on_vector, off_vector=off_vector1, aeff=aeff, edisp=edisp
    )
    obs2 = SpectrumObservation(
        on_vector=on_vector, off_vector=off_vector2, aeff=aeff, edisp=edisp
    )

    obs_list = [obs1, obs2]
    return obs_list


class SpectrumObservationTester:
    def __init__(self, obs, vals):
        self.obs = obs
        self.vals = vals

    def test_all(self):
        self.test_basic()
        self.test_stats_table()
        self.test_total_stats()
        self.test_stats_in_safe_range()
        self.test_to_sherpa()
        self.test_peek()
        self.test_npred()

    def test_basic(self):
        assert "Observation summary report" in str(self.obs)

    def test_npred(self):
        pwl = models.PowerLaw(
            index=2, amplitude=2e-11 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
        )
        npred = self.obs.predicted_counts(model=pwl)
        assert_allclose(npred.total_counts.value, self.vals["npred"], rtol=1e-3)

    def test_stats_table(self):
        table = self.obs.stats_table()
        assert table["n_on"].sum() == self.vals["total_on"]
        assert_quantity_allclose(
            table["livetime"].quantity.max(), self.vals["livetime"]
        )

    def test_total_stats(self):
        excess = self.obs.total_stats.excess
        assert_allclose(excess, self.vals["excess"], atol=1e-3)

    def test_stats_in_safe_range(self):
        stats = self.obs.total_stats_safe_range
        assert_quantity_allclose(stats.energy_min, self.obs.lo_threshold)
        assert_quantity_allclose(stats.energy_max, self.obs.hi_threshold)
        assert_allclose(stats.excess, self.vals["excess_safe_range"], rtol=1e-3)

    @requires_dependency("sherpa")
    def test_to_sherpa(self):
        # This method is not used anywhere but could be useful in the future
        sherpa_obs = self.obs.to_sherpa()
        assert sherpa_obs.counts[10] == self.obs.on_vector.data.data[10].value

    @requires_dependency("matplotlib")
    def test_peek(self):
        with mpl_plot_check():
            self.obs.peek()


def _read_hess_obs():
    path = "$GAMMAPY_DATA/joint-crab/spectra/hess/"
    obs1 = SpectrumObservation.read(path + "pha_obs23523.fits")
    obs2 = SpectrumObservation.read(path + "pha_obs23592.fits")
    return SpectrumObservationList([obs1, obs2])


@requires_data("gammapy-data")
class TestSpectrumObservationStacker:
    def setup(self):
        self.obs_list = _read_hess_obs()

        # Change threshold to make stuff more interesting
        self.obs_list.obs(23523).lo_threshold = 1.2 * u.TeV
        self.obs_list.obs(23523).hi_threshold = 50 * u.TeV
        self.obs_list.obs(23592).hi_threshold = 20 * u.TeV
        self.obs_stacker = SpectrumObservationStacker(self.obs_list)
        self.obs_stacker.run()

    def test_basic(self):
        assert "Stacker" in str(self.obs_stacker)
        assert "stacked" in str(self.obs_stacker.stacked_obs.on_vector.phafile)
        counts1 = self.obs_list[0].total_stats_safe_range.n_on
        counts2 = self.obs_list[1].total_stats_safe_range.n_on
        summed_counts = counts1 + counts2
        stacked_counts = self.obs_stacker.stacked_obs.total_stats.n_on
        assert summed_counts == stacked_counts

    def test_thresholds(self):
        energy = self.obs_stacker.stacked_obs.lo_threshold
        assert energy.unit == "keV"
        assert_allclose(energy.value, 8.912509e08, rtol=1e-3)

        energy = self.obs_stacker.stacked_obs.hi_threshold
        assert energy.unit == "keV"
        assert_allclose(energy.value, 4.466836e10, rtol=1e-3)

    def test_verify_npred(self):
        """Veryfing npred is preserved during the stacking"""
        pwl = models.PowerLaw(
            index=2, amplitude=2e-11 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
        )

        npred_stacked = self.obs_stacker.stacked_obs.predicted_counts(model=pwl)

        npred1 = self.obs_list[0].predicted_counts(model=pwl)
        npred2 = self.obs_list[1].predicted_counts(model=pwl)
        # Set npred outside safe range to 0
        npred1.data.data[np.nonzero(self.obs_list[0].on_vector.quality)] = 0
        npred2.data.data[np.nonzero(self.obs_list[1].on_vector.quality)] = 0

        npred_summed = npred1.data.data + npred2.data.data

        assert_allclose(npred_stacked.data.data, npred_summed)

    def test_stack_backscal(self):
        """Verify backscal stacking """
        obs_list = make_observation_list()
        obs_stacker = SpectrumObservationStacker(obs_list)
        obs_stacker.run()
        assert_allclose(obs_stacker.stacked_obs.alpha[0], 1.25 / 4.0)
        # When the OFF stack observation counts=0, the alpha is averaged on the total OFF counts for each run.
        assert_allclose(obs_stacker.stacked_obs.alpha[1], 2.5 / 8.0)


@requires_data("gammapy-data")
class TestSpectrumObservationList:
    def setup(self):
        self.obs_list = _read_hess_obs()

    def test_stack_method(self):
        obs = self.obs_list.stack()
        assert "Observation summary report" in str(obs)
        assert obs.obs_id == [23523, 23592]

        val = obs.aeff.data.evaluate(energy="1.1 TeV")
        assert val.unit == "cm2"
        assert_allclose(val.value, 1.34389e09, rtol=1e-3)

        val = obs.edisp.data.evaluate(e_true="1.1 TeV", e_reco="1.3 TeV")
        assert val.unit == ""
        assert_allclose(val.value, 0.075221, rtol=1e-3)

    def test_write(self, tmpdir):
        self.obs_list.write(outdir=str(tmpdir), pha_typeII=False)
        written_files = make_path(tmpdir).glob("*")
        assert len(list(written_files)) == len(self.obs_list) * 4

        outdir = tmpdir / "pha_typeII"
        self.obs_list.write(outdir=str(outdir), pha_typeII=True)

        test_list = SpectrumObservationList.read(outdir, pha_typeII=True)
        assert str(test_list[0].total_stats) == str(self.obs_list[0].total_stats)

    def test_range(self):
        energy = self.obs_list.safe_range("inclusive")
        assert energy.unit == "TeV"
        assert_allclose(energy.value, [0.891251, 100], rtol=1e-3)

        # TODO: this is not a great test case, should pick two
        # observations where "exclusive" and "inclusive" ranges differ.
        energy = self.obs_list.safe_range("exclusive")
        assert_allclose(energy.value, [0.891251, 100], rtol=1e-3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import astropy.units as u
from astropy.coordinates import Angle
from ..irf import EnergyDependentTablePSF
from ..maps import Map
from ..cube import PSFKernel

__all__ = ["make_psf_map", "PSFMap"]


def make_psf_map(psf, pointing, geom, max_offset):
    """Make a psf map for a single observation

    Expected axes : rad and true energy in this specific order
    The name of the rad MapAxis is expected to be 'rad'

    Parameters
    ----------
    psf : `~gammapy.irf.PSF3D`
        the PSF IRF
    pointing : `~astropy.coordinates.SkyCoord`
        the pointing direction
    geom : `~gammapy.maps.MapGeom`
        the map geom to be used. It provides the target geometry.
        rad and true energy axes should be given in this specific order.
    max_offset : `~astropy.coordinates.Angle`
        maximum offset w.r.t. fov center

    Returns
    -------
    psfmap : `~gammapy.cube.PSFMap`
        the resulting PSF map
    """
    energy_axis = geom.get_axis_by_name("energy")
    energy = energy_axis.center * energy_axis.unit

    rad_axis = geom.get_axis_by_name("theta")
    rad = Angle(rad_axis.center, unit=rad_axis.unit)

    # Compute separations with pointing position
    separations = pointing.separation(geom.to_image().get_coord().skycoord)
    valid = np.where(separations < max_offset)

    # Compute PSF values
    psf_values = psf.evaluate(offset=separations[valid], energy=energy, rad=rad)

    # Re-order axes to be consistent with expected geometry
    psf_values = np.transpose(psf_values, axes=(2, 0, 1))

    # Create Map and fill relevant entries
    psfmap = Map.from_geom(geom, unit="sr-1")
    psfmap.data[:, :, valid[0], valid[1]] += psf_values.to_value(psfmap.unit)

    return PSFMap(psfmap)


class PSFMap(object):
    """Class containing the Map of PSFs and allowing to interact with it.

    Parameters
    ----------
    psf_map : `~gammapy.maps.Map`
        the input PSF Map. Should be a Map with 2 non spatial axes.
        rad and true energy axes should be given in this specific order.

    Examples
    --------
    ::

        import numpy as np
        from astropy import units as u
        from astropy.coordinates import SkyCoord
        from gammapy.maps import Map, WcsGeom, MapAxis
        from gammapy.irf import EnergyDependentMultiGaussPSF
        from gammapy.cube import make_psf_map, PSFMap

        # Define energy axis. Note that the name is fixed.
        energy_axis = MapAxis.from_edges(np.logspace(-1., 1., 4), unit='TeV', name='energy')
        # Define rad axis. Again note the axis name
        rads = np.linspace(0., 0.5, 100) * u.deg
        rad_axis = MapAxis.from_edges(rads, unit='deg', name='theta')

        # Define parameters
        pointing = SkyCoord(0., 0., unit='deg')
        max_offset = 4 * u.deg

        # Create WcsGeom
        geom = WcsGeom.create(binsz=0.25*u.deg, width=10*u.deg, skydir=pointing, axes=[rad_axis, energy_axis])

        # Extract EnergyDependentTablePSF from CTA 1DC IRF
        filename = '$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits'
        psf = EnergyDependentMultiGaussPSF.read(filename, hdu='POINT SPREAD FUNCTION')
        psf3d = psf.to_psf3d(rads)

        # create the PSFMap for the specified pointing
        psf_map = make_psf_map(psf3d, pointing, geom, max_offset)

        # Get an EnergyDependentTablePSF at any position in the image
        psf_table = psf_map.get_energy_dependent_table_psf(SkyCoord(2., 2.5, unit='deg'))

        # Write map to disk
        psf_map.write('psf_map.fits')
    """

    def __init__(self, psf_map):
        if psf_map.geom.axes[1].name.upper() != "ENERGY":
            raise ValueError("Incorrect energy axis position in input Map")

        if psf_map.geom.axes[0].name.upper() != "THETA":
            raise ValueError("Incorrect theta axis position in input Map")

        self._psf_map = psf_map

    @property
    def psf_map(self):
        """the PSFMap itself (`~gammapy.maps.Map`)"""
        return self._psf_map

    @property
    def data(self):
        """the PSFMap data"""
        return self._psf_map.data

    @property
    def quantity(self):
        """the PSFMap data as a quantity"""
        return self._psf_map.quantity

    @property
    def geom(self):
        """The PSFMap MapGeom object"""
        return self._psf_map.geom

    @classmethod
    def read(cls, filename, **kwargs):
        """Read a psf_map from file and create a PSFMap object"""
        psfmap = Map.read(filename, **kwargs)
        return cls(psfmap)

    def write(self, *args, **kwargs):
        """Write the Map object containing the PSF Library map."""
        self._psf_map.write(*args, **kwargs)

    def get_energy_dependent_table_psf(self, position):
        """ Returns EnergyDependentTable PSF at a given position

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            the target position. Should be a single coordinates

        Returns
        -------
        psf_table : `~gammapy.irf.EnergyDependentTablePSF`
            the table PSF
        """
        if position.size != 1:
            raise ValueError(
                "EnergyDependentTablePSF can be extracted at one single position only."
            )

        # axes ordering fixed. Could be changed.
        pix_ener = np.arange(self.geom.axes[1].nbin)
        pix_rad = np.arange(self.geom.axes[0].nbin)

        # Convert position to pixels
        pix_lon, pix_lat = self.psf_map.geom.to_image().coord_to_pix(position)

        # Build the pixels tuple
        pix = np.meshgrid(pix_lon, pix_lat, pix_rad, pix_ener)

        # Interpolate in the PSF map. Squeeze to remove dimensions of length 1
        psf_values = np.squeeze(
            self.psf_map.interp_by_pix(pix) * u.Unit(self.psf_map.unit)
        )

        energies = self.psf_map.geom.axes[1].center * self.psf_map.geom.axes[1].unit
        rad = self.psf_map.geom.axes[0].center * self.psf_map.geom.axes[0].unit

        # Beware. Need to revert rad and energies to follow the TablePSF scheme.
        return EnergyDependentTablePSF(energy=energies, rad=rad, psf_value=psf_values.T)

    def get_psf_kernel(self, position, geom, max_radius=None, factor=4):
        """Returns a PSF kernel at the given position.

        The PSF is returned in the form a WcsNDMap defined by the input MapGeom.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            the target position. Should be a single coordinate
        geom : `~gammapy.maps.MapGeom`
            the target geometry to use
        max_radius : `~astropy.coordinates.Angle`
            maximum angular size of the kernel map
        factor : int
            oversampling factor to compute the PSF

        Returns
        -------
        kernel : `~gammapy.cube.PSFKernel`
            the resulting kernel
        """
        table_psf = self.get_energy_dependent_table_psf(position)
        return PSFKernel.from_table_psf(table_psf, geom, max_radius, factor)

    def containment_radius_map(self, energy, fraction=0.68):
        """Containment radius map.

        Parameters
        ----------
        energy : `~astropy.units.Quantity`
            Scalar energy at which to compute the containment radius
        fraction : float
            the containment fraction (range: 0 to 1)

        Returns
        -------
        containment_radius_map : `~gammapy.maps.Map`
            Containment radius map
        """
        coords = self.geom.to_image().get_coord().skycoord.flatten()
        m = Map.from_geom(self.geom.to_image(), unit="deg")

        for coord in coords:
            psf_table = self.get_energy_dependent_table_psf(coord)
            containment_radius = psf_table.containment_radius(energy, fraction)
            m.fill_by_coord(coord, containment_radius)

        return m
